# This module is renamed into preprocessing. Please use the preprocessing
# module instead of this module.

import warnings

from recordlinkage.preprocessing import clean as _clean
from recordlinkage.preprocessing import phonenumbers as _phonenumbers
from recordlinkage.preprocessing import value_occurence as _value_occurence
from recordlinkage.preprocessing import phonetic as _phonetic

from recordlinkage.utils import VisibleDeprecationWarning


def _depr_warn():

    warnings.warn(
        "module recordlinkage.standardise is deprecated, use "
        "recordlinkage.preprocessing instead",
        VisibleDeprecationWarning, stacklevel=2
    )


def clean(*args, **kwargs):

    _depr_warn()

    return _clean(*args, **kwargs)


def phonenumbers(*args, **kwargs):

    _depr_warn()

    return _phonenumbers(*args, **kwargs)


def value_occurence(*args, **kwargs):

    _depr_warn()

    return _value_occurence(*args, **kwargs)


def phonetic(*args, **kwargs):

    _depr_warn()

    return _phonetic(*args, **kwargs)



from __future__ import absolute_import, division, print_function

import recordlinkage as rl
from recordlinkage.datasets import load_febrl1, load_febrl4


class CompareRecordLinkage(object):

    timeout = 30 * 60

    def setup(self):

        # download data
        self.A, self.B = load_febrl4()

        # make pairs
        c_pairs = rl.FullIndex()
        pairs = c_pairs.index(self.A, self.B)

        # different sizes of pairs
        self.pairs_xsmall = pairs[0:5e3]
        self.pairs_small = pairs[0:5e4]
        self.pairs_medium = pairs[0:5e5]
        self.pairs_large = pairs[0:5e6]

    def time_global_xsmall(self):

        c_compare = rl.Compare(self.pairs_xsmall, self.A, self.B)
        c_compare.string('given_name', 'given_name', method='jaro')
        c_compare.string('surname', 'surname',
                         method='jarowinkler', threshold=0.85)
        c_compare.date('date_of_birth', 'date_of_birth')
        c_compare.exact('suburb', 'suburb')
        c_compare.exact('state', 'state')
        c_compare.string('address_1', 'address_1',
                         method='levenshtein', threshold=0.85)

    def time_global_small(self):

        c_compare = rl.Compare(self.pairs_small, self.A, self.B)
        c_compare.string('given_name', 'given_name', method='jaro')
        c_compare.string('surname', 'surname',
                         method='jarowinkler', threshold=0.85)
        c_compare.date('date_of_birth', 'date_of_birth')
        c_compare.exact('suburb', 'suburb')
        c_compare.exact('state', 'state')
        c_compare.string('address_1', 'address_1',
                         method='levenshtein', threshold=0.85)

    def time_global_medium(self):

        c_compare = rl.Compare(self.pairs_medium, self.A, self.B)
        c_compare.string('given_name', 'given_name', method='jaro')
        c_compare.string('surname', 'surname',
                         method='jarowinkler', threshold=0.85)
        c_compare.date('date_of_birth', 'date_of_birth')
        c_compare.exact('suburb', 'suburb')
        c_compare.exact('state', 'state')
        c_compare.string('address_1', 'address_1',
                         method='levenshtein', threshold=0.85)

    def time_global_large(self):

        c_compare = rl.Compare(self.pairs_large, self.A, self.B)
        c_compare.string('given_name', 'given_name', method='jaro')
        c_compare.string('surname', 'surname',
                         method='jarowinkler', threshold=0.85)
        c_compare.date('date_of_birth', 'date_of_birth')
        c_compare.exact('suburb', 'suburb')
        c_compare.exact('state', 'state')
        c_compare.string('address_1', 'address_1',
                         method='levenshtein', threshold=0.85)


class CompareDeduplication(object):

    timeout = 30 * 60

    def setup(self):

        # download data
        self.A = load_febrl1()

        # make pairs
        c_pairs = rl.FullIndex()
        pairs = c_pairs.index(self.A)

        # different sizes of pairs
        self.pairs_xsmall = pairs[0:5e3]
        self.pairs_small = pairs[0:5e4]
        self.pairs_medium = pairs[0:5e5]
        self.pairs_large = pairs[0:5e6]

    def time_global_xsmall(self):

        c_compare = rl.Compare(self.pairs_xsmall, self.A)
        c_compare.string('given_name', 'given_name', method='jaro')
        c_compare.string('surname', 'surname',
                         method='jarowinkler', threshold=0.85)
        c_compare.date('date_of_birth', 'date_of_birth')
        c_compare.exact('suburb', 'suburb')
        c_compare.exact('state', 'state')
        c_compare.string('address_1', 'address_1',
                         method='levenshtein', threshold=0.85)

    def time_global_small(self):

        c_compare = rl.Compare(self.pairs_small, self.A)
        c_compare.string('given_name', 'given_name', method='jaro')
        c_compare.string('surname', 'surname',
                         method='jarowinkler', threshold=0.85)
        c_compare.date('date_of_birth', 'date_of_birth')
        c_compare.exact('suburb', 'suburb')
        c_compare.exact('state', 'state')
        c_compare.string('address_1', 'address_1',
                         method='levenshtein', threshold=0.85)

    def time_global_medium(self):

        c_compare = rl.Compare(self.pairs_medium, self.A)
        c_compare.string('given_name', 'given_name', method='jaro')
        c_compare.string('surname', 'surname',
                         method='jarowinkler', threshold=0.85)
        c_compare.date('date_of_birth', 'date_of_birth')
        c_compare.exact('suburb', 'suburb')
        c_compare.exact('state', 'state')
        c_compare.string('address_1', 'address_1',
                         method='levenshtein', threshold=0.85)

    def time_global_large(self):

        c_compare = rl.Compare(self.pairs_large, self.A)
        c_compare.string('given_name', 'given_name', method='jaro')
        c_compare.string('surname', 'surname',
                         method='jarowinkler', threshold=0.85)
        c_compare.date('date_of_birth', 'date_of_birth')
        c_compare.exact('suburb', 'suburb')
        c_compare.exact('state', 'state')
        c_compare.string('address_1', 'address_1',
                         method='levenshtein', threshold=0.85)


class CompareAlgorithms(object):

    timeout = 30 * 60

    def setup(self):

        # download data
        self.A, self.B = load_febrl4()

        # Add numbers (age)
        self.A['postcode'] = self.A['postcode'].astype(float)
        self.B['postcode'] = self.B['postcode'].astype(float)

        # make pairs
        c_pairs = rl.FullIndex()
        self.pairs = c_pairs.index(self.A, self.B)[0:5e4]

# ************* STRING *************

    def time_string_jaro(self):

        c_compare = rl.Compare(self.pairs, self.A, self.B)
        c_compare.string('given_name', 'given_name', method='jaro')

    def time_string_jarowinkler(self):

        c_compare = rl.Compare(self.pairs, self.A, self.B)
        c_compare.string('given_name', 'given_name', method='jarowinkler')

    def time_string_qgram(self):

        c_compare = rl.Compare(self.pairs, self.A, self.B)
        c_compare.string('given_name', 'given_name', method='qgram')

    def time_string_cosine(self):

        c_compare = rl.Compare(self.pairs, self.A, self.B)
        c_compare.string('given_name', 'given_name', method='cosine')

    def time_string_levenshtein(self):

        c_compare = rl.Compare(self.pairs, self.A, self.B)
        c_compare.string('given_name', 'given_name', method='levenshtein')

# ************* Exact *************

    def time_exact(self):

        c_compare = rl.Compare(self.pairs, self.A, self.B)
        c_compare.exact('state', 'state')

# ************* NUMERIC *************

    def time_numeric_gauss(self):

        c_compare = rl.Compare(self.pairs, self.A, self.B)
        c_compare.numeric('age', 'age', method='gauss', scale=2)



# measures.py

from __future__ import division

import numpy
import pandas

from recordlinkage.utils import get_length


def _get_multiindex(x):

    if isinstance(x, (pandas.DataFrame, pandas.Series)):
        return x.index
    elif isinstance(x, pandas.MultiIndex):
        return x
    else:
        raise ValueError("Expected one of: pandas.DataFrame, "
                         "pandas.Series, pandas.MultiIndex")


def _isconfusionmatrix(x):

    if isinstance(x, numpy.ndarray) and x.shape == (2, 2):
        return True
    elif isinstance(x, list) and numpy.ndarray(x).shape == (2, 2):
        return True
    else:
        return False


def reduction_ratio(links_pred, *total):
    """Compute the reduction ratio.

    The reduction ratio is 1 minus the ratio candidate matches and the maximum
    number of pairs possible.

    Parameters
    ----------
    links_pred: int, pandas.MultiIndex
        The number of candidate record pairs or the pandas.MultiIndex with
        record pairs.
    *total: pandas.DataFrame object(s)
        The DataFrames are used to compute the full index size with the
        full_index_size function.

    Returns
    -------
    float
        The reduction ratio.

    """

    n_max = full_index_size(*total)

    if isinstance(links_pred, pandas.MultiIndex):
        links_pred = len(links_pred)

    if links_pred > n_max:
        raise ValueError("n has to be smaller of equal n_max")

    return 1 - links_pred / n_max


def max_pairs(shape):
    """[DEPRECATED] Compute the maximum number of record pairs possible."""

    if not isinstance(shape, (tuple, list)):
        x = get_length(shape)
        n = int(x * (x - 1) / 2)

    elif (isinstance(shape, (tuple, list)) and len(shape) == 1):
        x = get_length(shape[0])
        n = int(x * (x - 1) / 2)

    else:
        n = numpy.prod([get_length(xi) for xi in shape])

    return n


def full_index_size(*args):
    """Compute the number of records in a full index.

    Compute the number of records in a full index without building the index
    itself. The result is the maximum number of record pairs possible. This
    function is especially useful in measures like the `reduction_ratio`.

    Deduplication: Given a DataFrame A with length N, the full index size is
    N*(N-1)/2. Linking: Given a DataFrame A with length N and a DataFrame B
    with length M, the full index size is N*M.

    Parameters
    ----------
    *args: int, pandas.MultiIndex, pandas.Series, pandas.DataFrame
        A pandas object or a int representing the length of a dataset to link.
        When there is one argument, it is assumed that the record linkage is
        a deduplication process.

    Examples
    --------

    Use integers:
    >>> full_index_size(10)  # deduplication: 45 pairs
    >>> full_index_size(10, 10)  # linking: 100 pairs

    or pandas objects
    >>> full_index_size(DF)  # deduplication: len(DF)*(len(DF)-1)/2 pairs
    >>> full_index_size(DF, DF)  # linking: len(DF)*len(DF) pairs

    """

    # check if a list or tuple is passed as argument
    if len(args) == 1 and isinstance(args[0], (list, tuple)):
        args = tuple(args[0])

    if len(args) == 1:
        n = get_length(args[0])
        size = int(n * (n - 1) / 2)
    else:
        size = numpy.prod([get_length(arg) for arg in args])

    return size


def true_positives(links_true, links_pred):
    """Count the number of True Positives.

    Returns the number of correctly predicted links, also called the number of
    True Positives (TP).

    Parameters
    ----------
    links_true: pandas.MultiIndex, pandas.DataFrame, pandas.Series
        The true (or actual) links.
    links_pred: pandas.MultiIndex, pandas.DataFrame, pandas.Series
        The predicted links.

    Returns
    -------
    int
        The number of correctly predicted links.
    """

    links_true = _get_multiindex(links_true)
    links_pred = _get_multiindex(links_pred)

    return len(links_true & links_pred)


def true_negatives(links_true, links_pred, total):
    """Count the number of True Negatives.

    Returns the number of correctly predicted non-links, also called the
    number of True Negatives (TN).

    Parameters
    ----------
    links_true: pandas.MultiIndex, pandas.DataFrame, pandas.Series
        The true (or actual) links.
    links_pred: pandas.MultiIndex, pandas.DataFrame, pandas.Series
        The predicted links.
    total: int, pandas.MultiIndex
        The count of all record pairs (both links and non-links). When the
        argument is a pandas.MultiIndex, the length of the index is used.

    Returns
    -------
    int
        The number of correctly predicted non-links.

    """

    links_true = _get_multiindex(links_true)
    links_pred = _get_multiindex(links_pred)

    if isinstance(total, pandas.MultiIndex):
        total = len(total)

    return int(total) - len(links_true | links_pred)


def false_positives(links_true, links_pred):
    """Count the number of False Positives.

    Returns the number of incorrect predictions of true non-links. (true non-
    links, but predicted as links). This value is known as the number of False
    Positives (FP).

    Parameters
    ----------
    links_true: pandas.MultiIndex, pandas.DataFrame, pandas.Series
        The true (or actual) links.
    links_pred: pandas.MultiIndex, pandas.DataFrame, pandas.Series
        The predicted links.

    Returns
    -------
    int
        The number of false positives.

    """

    links_true = _get_multiindex(links_true)
    links_pred = _get_multiindex(links_pred)

    return len(links_pred.difference(links_true))


def false_negatives(links_true, links_pred):
    """Count the number of False Negatives.

    Returns the number of incorrect predictions of true links. (true links,
    but predicted as non-links). This value is known as the number of False
    Negatives (FN).

    Parameters
    ----------
    links_true: pandas.MultiIndex, pandas.DataFrame, pandas.Series
        The true (or actual) links.
    links_pred: pandas.MultiIndex, pandas.DataFrame, pandas.Series
        The predicted links.

    Returns
    -------
    int
        The number of false negatives.

    """

    links_true = _get_multiindex(links_true)
    links_pred = _get_multiindex(links_pred)

    return len(links_true.difference(links_pred))


def confusion_matrix(links_true, links_pred, total=None):
    """Compute the confusion matrix.

    The confusion matrix is of the following form:

    +----------------------+-----------------------+----------------------+
    |                      |  Predicted Positives  | Predicted Negatives  |
    +======================+=======================+======================+
    | **True Positives**   | True Positives (TP)   | False Negatives (FN) |
    +----------------------+-----------------------+----------------------+
    | **True Negatives**   | False Positives (FP)  | True Negatives (TN)  |
    +----------------------+-----------------------+----------------------+

    The confusion matrix is an informative way to analyse a prediction. The
    matrix can used to compute measures like precision and recall. The count
    of true prositives is [0,0], false negatives is [0,1], true negatives
    is [1,1] and false positives is [1,0].

    Parameters
    ----------
    links_true: pandas.MultiIndex, pandas.DataFrame, pandas.Series
        The true (or actual) links.
    links_pred: pandas.MultiIndex, pandas.DataFrame, pandas.Series
        The predicted links.
    total: int, pandas.MultiIndex
        The count of all record pairs (both links and non-links). When the
        argument is a pandas.MultiIndex, the length of the index is used. If
        the total is None, the number of True Negatives is not computed.
        Default None.

    Returns
    -------
    numpy.array
        The confusion matrix with TP, TN, FN, FP values.

    Note
    ----
    The number of True Negatives is computed based on the total argument.
    This argument is the number of record pairs of the entire matrix.

    """

    links_true = _get_multiindex(links_true)
    links_pred = _get_multiindex(links_pred)

    tp = true_positives(links_true, links_pred)
    fp = false_positives(links_true, links_pred)
    fn = false_negatives(links_true, links_pred)

    if total is None:
        tn = numpy.nan
    else:
        tn = true_negatives(links_true, links_pred, total)

    return numpy.array([[tp, fn], [fp, tn]])


def precision(links_true, links_pred=None):
    """precision(links_true, links_pred)

    Compute the precision.

    The precision is given by TP/(TP+FP).

    Parameters
    ----------
    links_true: pandas.MultiIndex, pandas.DataFrame, pandas.Series
        The true (or actual) collection of links.
    links_pred: pandas.MultiIndex, pandas.DataFrame, pandas.Series
        The predicted collection of links.

    Returns
    -------
    float
        The precision
    """

    if _isconfusionmatrix(links_true):

        confusion_matrix = links_true

        v = confusion_matrix[0, 0] \
            / (confusion_matrix[0, 0] + confusion_matrix[1, 0])
    else:

        tp = true_positives(links_true, links_pred)
        fp = false_positives(links_true, links_pred)
        v = tp / (tp + fp)

    return float(v)


def recall(links_true, links_pred=None):
    """recall(links_true, links_pred)

    Compute the recall/sensitivity.

    The recall is given by TP/(TP+FN).

    Parameters
    ----------
    links_true: pandas.MultiIndex, pandas.DataFrame, pandas.Series
        The true (or actual) collection of links.
    links_pred: pandas.MultiIndex, pandas.DataFrame, pandas.Series
        The predicted collection of links.

    Returns
    -------
    float
        The recall
    """

    if _isconfusionmatrix(links_true):

        confusion_matrix = links_true

        v = confusion_matrix[0, 0] \
            / (confusion_matrix[0, 0] + confusion_matrix[0, 1])
    else:

        tp = true_positives(links_true, links_pred)
        fn = false_negatives(links_true, links_pred)
        v = tp / (tp + fn)

    return float(v)


def accuracy(links_true, links_pred=None, total=None):
    """accuracy(links_true, links_pred, total)

    Compute the accuracy.

    The accuracy is given by (TP+TN)/(TP+FP+TN+FN).

    Parameters
    ----------
    links_true: pandas.MultiIndex, pandas.DataFrame, pandas.Series
        The true (or actual) collection of links.
    links_pred: pandas.MultiIndex, pandas.DataFrame, pandas.Series
        The predicted collection of links.
    total: int, pandas.MultiIndex
        The count of all record pairs (both links and non-links). When the
        argument is a pandas.MultiIndex, the length of the index is used.

    Returns
    -------
    float
        The accuracy
    """

    if _isconfusionmatrix(links_true):

        confusion_matrix = links_true

        v = (confusion_matrix[0, 0] + confusion_matrix[1, 1]) \
            / numpy.sum(confusion_matrix)
    else:

        tp = true_positives(links_true, links_pred)
        tn = true_negatives(links_true, links_pred, total)

        v = (tp + tn) / total

    return float(v)


def specificity(links_true, links_pred=None, total=None):
    """specificity(links_true, links_pred, total)

    Compute the specificity.

    The specificity is given by TN/(FP+TN).

    Parameters
    ----------
    links_true: pandas.MultiIndex, pandas.DataFrame, pandas.Series
        The true (or actual) collection of links.
    links_pred: pandas.MultiIndex, pandas.DataFrame, pandas.Series
        The predicted collection of links.
    total: int, pandas.MultiIndex
        The count of all record pairs (both links and non-links). When the
        argument is a pandas.MultiIndex, the length of the index is used.

    Returns
    -------
    float
        The specificity

    """

    if _isconfusionmatrix(links_true):

        confusion_matrix = links_true

        v = confusion_matrix[1, 1] / \
            (confusion_matrix[1, 0] + confusion_matrix[1, 1])
    else:

        fp = false_positives(links_true, links_pred)
        tn = true_negatives(links_true, links_pred, total)
        v = tn / (fp + tn)

    return float(v)


def fscore(links_true, links_pred=None):
    """fscore(links_true, links_pred)

    Compute the F-score.

    The F-score is given by 2*(precision*recall)/(precision+recall).

    Parameters
    ----------
    links_true: pandas.MultiIndex, pandas.DataFrame, pandas.Series
        The true (or actual) collection of links.
    links_pred: pandas.MultiIndex, pandas.DataFrame, pandas.Series
        The predicted collection of links.

    Returns
    -------
    float
        The fscore

    Note
    ----
    If there are no pairs predicted as links, this measure will raise a
    ZeroDivisionError.

     """

    prec = precision(links_true, links_pred)
    rec = recall(links_true, links_pred)

    return float(2 * prec * rec / (prec + rec))



from recordlinkage.indexing import *
from recordlinkage.comparing import *
from recordlinkage.classifiers import *
from recordlinkage.measures import *
from recordlinkage.utils import split_index

from recordlinkage import rl_logging as logging

from recordlinkage._version import get_versions
__version__ = get_versions()['version']
del get_versions

__all__ = ['datasets', 'preprocessing']



import pandas
import numpy as np


# Numerical comparison algorithms

def _step_sim(d, offset=0, origin=0):
    # scale is not an argument

    if offset < 0:
        raise ValueError("The offset must be positive.")

    expr = 'abs(d - origin) <= offset'

    return pandas.eval(expr).astype(np.float64)


def _linear_sim(d, scale, offset=0, origin=0):

    if offset < 0:
        raise ValueError("The offset must be positive.")

    if scale <= 0:
        raise ValueError("The scale must be larger than 0. ")

    d = (abs(d - origin)).clip(offset, offset + 2 * scale)

    expr = '1 - (d-offset)/(2*scale)'

    return pandas.eval(expr)


def _squared_sim(d, scale, offset=0, origin=0):

    if offset < 0:
        raise ValueError("The offset must be positive.")

    if scale <= 0:
        raise ValueError("The scale must be larger than 0. ")

    d = (abs(d - origin)).clip(offset, offset + np.sqrt(2) * scale)
    # solve y=1-ad^2 given y(d=scale)=0.5
    # 1-y = ad^2
    # a = (1-y)/d^2

    # fill y=0.5 and d = scale
    # a = (1-0.5)/scale^2
    # a = 1/(2*scale^2)
    # y = 1 - 1/2*(d/scale)^2
    # d = sqrt(2)*scale is the point where similarity is zero.

    expr = '1 - 1/2*exp(2*log((d-offset)/scale))'

    return pandas.eval(expr)


def _exp_sim(d, scale, offset=0, origin=0):

    if offset < 0:
        raise ValueError("The offset must be positive.")

    if scale <= 0:
        raise ValueError("The scale must be larger than 0. ")

    d = (abs(d - origin)).clip(offset, None)

    # solve y=exp(-x*a) if 1/2 = exp(-x/scale)
    expr = '2**(-(d-offset)/scale)'

    return pandas.eval(expr)


def _gauss_sim(d, scale, offset=0, origin=0):

    if offset < 0:
        raise ValueError("The offset must be positive.")

    if scale <= 0:
        raise ValueError("The scale must be larger than 0. ")

    d = (abs(d - origin)).clip(offset, None)

    # solve y=exp(-x^2*a) if 1/2 = exp(-x^2/scale^2)
    expr = '2**(-((d-offset)/scale)**2)'

    return pandas.eval(expr)



import warnings
import time

import pandas
import numpy

from sklearn import cluster, linear_model, naive_bayes, svm
from sklearn.utils.validation import NotFittedError

from recordlinkage.algorithms.em import ECMEstimate

from recordlinkage import rl_logging as logging


class LearningError(Exception):
    """Learning error"""


class Classifier(object):
    """Base class for classification of records pairs.

    This class contains methods for training the classifier. Distinguish
    different types of training, such as supervised and unsupervised learning.

    """

    def __init__(self):

        logging.info("Classification - initialize {} class".format(
            self.__class__.__name__)
        )

        # The actual classifier. Maybe this is slightly strange because of
        # inheritance.
        self.classifier = None

    def learn(self, comparison_vectors, match_index, return_type='index'):
        """Train the classifier.

        Parameters
        ----------
        comparison_vectors : pandas.DataFrame
            The comparison vectors.
        match_index : pandas.MultiIndex
            The true matches.
        return_type : 'index' (default), 'series', 'array'
            The format to return the classification result. The argument value
            'index' will return the pandas.MultiIndex of the matches. The
            argument value 'series' will return a pandas.Series with zeros
            (distinct) and ones (matches). The argument value 'array' will
            return a numpy.ndarray with zeros and ones.

        Returns
        -------
        pandas.Series
            A pandas Series with the labels 1 (for the matches) and 0 (for the
            non-matches).

        """

        logging.info("Classifying - start learning {}".format(
            self.__class__.__name__)
        )

        # start timing
        start_time = time.time()

        if isinstance(match_index, (pandas.MultiIndex, pandas.Index)):

            # The match_index variable is of type MultiIndex
            train_series = pandas.Series(False, index=comparison_vectors.index)

            try:
                train_series.loc[match_index & comparison_vectors.index] = True

            except pandas.IndexError as err:

                # The are no matches. So training is not possible.
                if len(match_index & comparison_vectors.index) == 0:
                    raise LearningError(
                        "both matches and non-matches needed in the" +
                        "trainingsdata, only non-matches found"
                    )
                else:
                    raise err

        self.classifier.fit(
            comparison_vectors.as_matrix(),
            numpy.array(train_series)
        )

        result = self._predict(comparison_vectors, return_type)

        # log timing
        logf_time = "Classifying - learning computation time: ~{:.2f}s"
        logging.info(logf_time.format(time.time() - start_time))

        return result

    def predict(self, comparison_vectors, return_type='index'):
        """Predict the class of the record pairs.

        Classify a set of record pairs based on their comparison vectors into
        matches, non-matches and possible matches. The classifier has to be
        trained to call this method.


        Parameters
        ----------
        comparison_vectors : pandas.DataFrame
            Dataframe with comparison vectors.
        return_type : 'index' (default), 'series', 'array'
            The format to return the classification result. The argument value
            'index' will return the pandas.MultiIndex of the matches. The
            argument value 'series' will return a pandas.Series with zeros
            (distinct) and ones (matches). The argument value 'array' will
            return a numpy.ndarray with zeros and ones.

        Returns
        -------
        pandas.Series
            A pandas Series with the labels 1 (for the matches) and 0 (for the
            non-matches).

        """

        logging.info("Classifying - predict matches and non-matches")

        return self._predict(comparison_vectors, return_type)

    def _predict(self, comparison_vectors, return_type):

        try:
            prediction = self.classifier.predict(
                comparison_vectors.as_matrix())
        except NotFittedError:
            raise NotFittedError(
                "This {} is not fitted yet. Call 'learn' with appropriate "
                "arguments before using this method.".format(
                    type(self).__name__
                )
            )

        return self._return_result(prediction, return_type, comparison_vectors)

    def prob(self, comparison_vectors, return_type='series'):
        """Compute the probabilities for each record pair.

        For each pair of records, estimate the probability of being a match.

        Parameters
        ----------
        comparison_vectors : pandas.DataFrame
            The dataframe with comparison vectors.
        return_type : 'series' or 'array'
            Return a pandas series or numpy array. Default 'series'.

        Returns
        -------
        pandas.Series or numpy.ndarray
            The probability of being a match for each record pair.

        """

        logging.info("Classifying - compute probabilities")

        probs = self.classifier.predict_proba(comparison_vectors.as_matrix())

        if return_type == 'series':
            return pandas.Series(probs[:, 0], index=comparison_vectors.index)
        elif return_type == 'array':
            return probs[:, 0]
        else:
            raise ValueError(
                "return_type {} unknown. Choose 'index', 'series' or "
                "'array'".format(return_type))

    def _return_result(
        self, result, return_type='index', comparison_vectors=None
    ):
        """Return different formatted classification results.

        """

        if type(result) != numpy.ndarray:
            raise ValueError("numpy.ndarray expected.")

        # return the pandas.MultiIndex
        if return_type == 'index':
            return comparison_vectors.index[result.astype(bool)]

        # return a pandas.Series
        elif return_type == 'series':
            return pandas.Series(
                result,
                index=comparison_vectors.index,
                name='classification')

        # return a numpy.ndarray
        elif return_type == 'array':
            return result

        # return_type not known
        else:
            raise ValueError(
                "return_type {} unknown. Choose 'index', 'series' or "
                "'array'".format(return_type))


class KMeansClassifier(Classifier):
    """KMeans classifier.

    The `K-means clusterings algorithm (wikipedia)
    <https://en.wikipedia.org/wiki/K-means_clustering>`_ partitions candidate
    record pairs into matches and non-matches. Each comparison vector belongs
    to the cluster with the nearest mean. The K-means algorithm does not need
    trainings data, but it needs two starting points (one for the matches and
    one for the non-matches). The K-means clustering problem is NP-hard.

    Parameters
    ----------
    match_cluster_center : list, numpy.array
        The center of the match cluster. The length of the list/array must
        equal the number of comparison variables.
    nonmatch_cluster_center : list, numpy.array
        The center of the nonmatch (distinct) cluster. The length of the
        list/array must equal the number of comparison variables.

    Attributes
    ----------
    classifier: sklearn.cluster.KMeans
        The Kmeans cluster class in sklearn.
    match_cluster_center : list, numpy.array
        The center of the match cluster.
    nonmatch_cluster_center : list, numpy.array
        The center of the nonmatch (distinct) cluster.

    Note
    ----
    There are way better methods for linking records than the k-means
    clustering algorithm. However, this algorithm does not need trainings data
    and is useful to do an initial partition.
    """

    def __init__(self, match_cluster_center=None,
                 nonmatch_cluster_center=None, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)

        # initialize the classifier
        self.classifier = cluster.KMeans(n_clusters=2, n_init=1)

        self._match_cluster_center = match_cluster_center
        self._nonmatch_cluster_center = nonmatch_cluster_center

        # set cluster centers if available
        self._set_cluster_centers(
            self._match_cluster_center, self._nonmatch_cluster_center)

    @property
    def match_cluster_center(self):
        # Return the centers if available
        try:
            return self.classifier.cluster_centers_.tolist()[1]
        except AttributeError:
            return None

    @match_cluster_center.setter
    def match_cluster_center(self, value):

        self._match_cluster_center = value

        try:
            self._set_cluster_centers(
                self._match_cluster_center, self._nonmatch_cluster_center)
        except ValueError:
            pass

    @property
    def nonmatch_cluster_center(self):
        # Return the centers if available
        try:
            return self.classifier.cluster_centers_.tolist()[0]
        except AttributeError:
            return None

    @nonmatch_cluster_center.setter
    def nonmatch_cluster_center(self, value):

        self._nonmatch_cluster_center = value

        try:
            self._set_cluster_centers(
                self._match_cluster_center, self._nonmatch_cluster_center)
        except ValueError:
            pass

    def _set_cluster_centers(self, m_cluster_center, n_cluster_center):

        if m_cluster_center is not None and n_cluster_center is not None:
            self.classifier.cluster_centers_ = numpy.array(
                [numpy.array(n_cluster_center), numpy.array(m_cluster_center)]
            )
        elif m_cluster_center is None and n_cluster_center is None:
            try:
                del self.classifier.cluster_centers_
            except AttributeError:
                pass
        else:
            raise ValueError("set the center of the match cluster and the " +
                             "nonmatch cluster")

    def learn(self, comparison_vectors, return_type='index'):
        """Train the K-means classifier.

        The K-means classifier is unsupervised and therefore does not need
        labels. The K-means classifier classifies the data into two sets of
        links and non- links. The starting point of the cluster centers are
        0.05 for the non-matches and 0.95 for the matches.

        Parameters
        ----------
        comparison_vectors : pandas.DataFrame
            The dataframe with comparison vectors.
        return_type : 'index' (default), 'series', 'array'
            The format to return the classification result. The argument value
            'index' will return the pandas.MultiIndex of the matches. The
            argument value 'series' will return a pandas.Series with zeros
            (distinct) and ones (matches). The argument value 'array' will
            return a numpy.ndarray with zeros and ones.

        Returns
        -------
        pandas.MultiIndex, pandas.Series or numpy.ndarray
            The prediction (see also the argument 'return_type')

        """

        # Set the start point of the classifier.
        self.classifier.init = numpy.array([
            [0.05] * len(list(comparison_vectors)),
            [0.95] * len(list(comparison_vectors))
        ])

        # Fit and predict
        prediction = self.classifier.fit_predict(
            comparison_vectors.as_matrix())

        return self._return_result(prediction, return_type, comparison_vectors)

    def prob(self, *args, **kwargs):

        raise AttributeError(
            "It is not possible to compute "
            "probabilities for the KMeansClassfier")


# DeterministicClassifier = LogisticRegressionClassifier
class LogisticRegressionClassifier(Classifier):
    """Logistic Regression Classifier.

    This classifier is an application of the `logistic regression model
    (wikipedia) <https://en.wikipedia.org/wiki/Logistic_regression>`_. The
    classifier partitions candidate record pairs into matches and non-matches.

    Parameters
    ----------
    coefficients : list, numpy.array
        The coefficients of the logistic regression.
    intercept : float
        The interception value.

    Attributes
    ----------
    classifier: sklearn.linear_model.LogisticRegression
        The Logistic regression classifier in sklearn.
    coefficients : list
        The coefficients of the logistic regression.
    intercept : float
        The interception value.

    """

    def __init__(self, coefficients=None, intercept=None):
        super(self.__class__, self).__init__()

        self.classifier = linear_model.LogisticRegression()

        self.coefficients = coefficients
        self.intercept = intercept

        self.classifier.classes_ = numpy.array([False, True])

    @property
    def params(self):
        return {
            'coefficients': self.coefficients,
            'intercept': self.intercept
        }

    @params.setter
    def params(self, value):

        if not isinstance(value, dict):
            raise ValueError("parameters are of wrong type")

        self.coefficients = value['coefficients']
        self.intercept = value['intercept']

    @property
    def coefficients(self):
        # Return the coefficients if available
        try:
            return list(self.classifier.coef_[0])
        except Exception:
            return None

    @coefficients.setter
    def coefficients(self, value):

        if value is not None:

            # Check if array if numpy array
            if type(value) is not numpy.ndarray:
                value = numpy.array(value)

            self.classifier.coef_ = value.reshape((1, len(value)))

    @property
    def intercept(self):

        try:
            return float(self.classifier.intercept_[0])
        except Exception:
            return None

    @intercept.setter
    def intercept(self, value):

        if value is not None:

            if not isinstance(value, (list)):
                value = numpy.array(value)
            else:
                value = numpy.array([value])

            self.classifier.intercept_ = value

        # value is None
        elif value is None:
            try:
                del self.classifier.intercept_
            except AttributeError:
                pass
        else:
            raise ValueError('incorrect type')


class NaiveBayesClassifier(Classifier):
    """
    NaiveBayesClassifier(alpha=1.0)

    Naive Bayes Classifier.

    The `Naive Bayes classifier (wikipedia)
    <https://en.wikipedia.org/wiki/Naive_Bayes_classifier>`_ partitions
    candidate record pairs into matches and non-matches. The classifier is
    based on probabilistic principles. The Naive Bayes classification method
    is proven to be mathematical equivalent with the Fellegi and Sunter model.

    Parameters
    ----------
    log_prior : list, numpy.array
        The log propabaility of each class.

    Attributes
    ----------
    classifier: sklearn.linear_model.LogisticRegression
        The Logistic regression classifier in sklearn.
    coefficients : list
        The coefficients of the logistic regression.
    intercept : float
        The interception value.

    Parameters
    ----------
    alpha : float
        Additive (Laplace/Lidstone) smoothing parameter (0 for no smoothing).

    """

    def __init__(self, log_prior=None, alpha=1.0, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)

        self.classifier = naive_bayes.BernoulliNB(alpha=alpha, binarize=None)

        self.log_prior = log_prior

    @property
    def log_prior(self):

        try:
            return self.classifier.class_log_prior_.tolist()
        except Exception:
            return None

    @log_prior.setter
    def log_prior(self, value):

        if isinstance(value, (list, numpy.ndarray)):

            self.classifier.class_log_prior_ = numpy.array(value)

        # value is None
        elif value is None:
            try:
                del self.classifier.class_log_prior_
            except AttributeError:
                pass
        else:
            raise ValueError('incorrect type')


class SVMClassifier(Classifier):
    """
    SVMClassifier()

    Support Vector Machines

    The `Support Vector Machine classifier (wikipedia)
    <https://en.wikipedia.org/wiki/Support_vector_machine>`_ partitions
    candidate record pairs into matches and non-matches. This implementation
    is a non-probabilistic binary linear classifier. Support vector machines
    are supervised learning models. Therefore, the SVM classifiers needs
    training-data.

    """

    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)

        self.classifier = svm.LinearSVC()

    def prob(self, *args, **kwargs):

        raise AttributeError(
            "It is not possible to compute "
            "probabilities for the SVMClassfier")


class FellegiSunter(Classifier):
    """Fellegi and Sunter framework.

    Base class for probabilistic classification of records pairs with the
    Fellegi and Sunter (1969) framework.

    """

    def __init__(self, random_decision_rule=False):

        self.random_decision_rule = random_decision_rule

    @property
    def p(self):
        try:
            return self.algorithm._p
        except Exception:
            pass

    def _decision_rule(self, probabilities, threshold,
                       random_decision_rule=False):

        if not self.random_decision_rule:
            return (probabilities >= threshold).astype(int)
        else:
            raise NotImplementedError(
                "Random decisions are not possible at the moment.")


class ECMClassifier(FellegiSunter):
    """Expectation/Conditional Maxisation vlassifier.

    [EXPERIMENTAL] Expectation/Conditional Maximisation algorithm used as
    classifier. This probabilistic record linkage algorithm is used in
    combination with Fellegi and Sunter model.

    """

    def __init__(self, *args, **kwargs):
        super(self.__class__, self).__init__(*args, **kwargs)

        self.algorithm = ECMEstimate()

    def learn(self, comparison_vectors, init='jaro', return_type='index'):
        """ Train the algorithm.

        Train the Expectation-Maximisation classifier. This method is well-
        known as the ECM-algorithm implementation in the context of record
        linkage.

        Parameters
        ----------
        comparison_vectors : pandas.DataFrame
            The dataframe with comparison vectors.
        params_init : dict
            A dictionary with initial parameters of the ECM algorithm
            (optional).
        return_type : 'index' (default), 'series', 'array'
            The format to return the classification result. The argument value
            'index' will return the pandas.MultiIndex of the matches. The
            argument value 'series' will return a pandas.Series with zeros
            (distinct) and ones (matches). The argument value 'array' will
            return a numpy.ndarray with zeros and ones.

        Returns
        -------
        pandas.Series
            A pandas Series with the labels 1 (for the matches) and 0 (for the
            non-matches).

        """

        logging.info("Classifying - start learning {}".format(
            self.__class__.__name__)
        )

        # start timing
        start_time = time.time()

        probs = self.algorithm.train(comparison_vectors.as_matrix())

        n_matches = int(self.algorithm.p * len(probs))
        self.p_threshold = numpy.sort(probs)[len(probs) - n_matches]

        prediction = self._decision_rule(probs, self.p_threshold)

        result = self._return_result(
            prediction, return_type, comparison_vectors
        )

        # log timing
        logf_time = "Classifying - learning computation time: ~{:.2f}s"
        logging.info(logf_time.format(time.time() - start_time))

        return result

    def predict(self, comparison_vectors, return_type='index',
                *args, **kwargs):
        """Predict the class of reord pairs.

        Classify a set of record pairs based on their comparison vectors into
        matches, non-matches and possible matches. The classifier has to be
        trained to call this method.

        Parameters
        ----------
        comparison_vectors : pandas.DataFrame
            The dataframe with comparison vectors.
        return_type : 'index' (default), 'series', 'array'
            The format to return the classification result. The argument value
            'index' will return the pandas.MultiIndex of the matches. The
            argument value 'series' will return a pandas.Series with zeros
            (distinct) and ones (matches). The argument value 'array' will
            return a numpy.ndarray with zeros and ones.

        Returns
        -------
        pandas.Series
            A pandas Series with the labels 1 (for the matches) and 0 (for the
            non-matches).

        Note
        ----
        Prediction is risky for this unsupervised learning method. Be aware
        that the sample from the population is valid.


        """

        logging.info("Classifying - predict matches and non-matches")

        enc_vectors = self.algorithm._transform_vectors(
            comparison_vectors.as_matrix())

        probs = self.algorithm._expectation(enc_vectors)

        prediction = self._decision_rule(probs, self.p_threshold)

        return self._return_result(prediction, return_type, comparison_vectors)

    def prob(self, comparison_vectors):
        """Compute the probabilities for each record pair.

        For each pair of records, estimate the probability of being a match.

        Parameters
        ----------
        comparison_vectors : pandas.DataFrame
            The dataframe with comparison vectors.
        return_type : 'series' or 'array'
            Return a pandas series or numpy array. Default 'series'.

        Returns
        -------
        pandas.Series or numpy.ndarray
            The probability of being a match for each record pair.

        """

        logging.info("Classifying - compute probabilities")

        enc_vectors = self.algorithm._transform_vectors(
            comparison_vectors.as_matrix())

        return pandas.Series(
            self.algorithm._expectation(enc_vectors),
            index=comparison_vectors.index
        )



from __future__ import absolute_import, division, print_function

import recordlinkage as rl
from recordlinkage.datasets import load_febrl1, load_febrl4


class PairsRecordLinkage(object):

    timeout = 30 * 60

    def setup(self):

        # download data
        self.A, self.B = load_febrl4()

    def time_full_index(self):

        # setup class
        c_pairs = rl.FullIndex()

        # Make pairs
        c_pairs.index(self.A, self.B)

    def time_block_index(self):

        # setup class
        c_pairs = rl.BlockIndex('given_name')

        # Make pairs
        c_pairs.index(self.A, self.B)

    def time_sni_index(self):

        # setup class
        c_pairs = rl.SortedNeighbourhoodIndex(on='given_name', w=5)

        # Make pairs
        c_pairs.index(self.A, self.B)

    def time_random_index(self):

        # setup class
        c_pairs = rl.RandomIndex(2500)

        # Make pairs
        c_pairs.index(self.A, self.B)


class PairsDeduplication(object):

    timeout = 30 * 60

    def setup(self):

        # download data
        self.A = load_febrl1()

    def time_full_index(self):

        # setup class
        c_pairs = rl.FullIndex()

        # Make pairs
        c_pairs.index(self.A)

    def time_block_index(self):

        # setup class
        c_pairs = rl.BlockIndex('given_name')

        # Make pairs
        c_pairs.index(self.A)

    def time_sni_index(self):

        # setup class
        c_pairs = rl.SortedNeighbourhoodIndex(on='given_name', w=5)

        # Make pairs
        c_pairs.index(self.A)

    def time_random_index(self):

        # setup class
        c_pairs = rl.RandomIndex(2500)

        # Make pairs
        c_pairs.index(self.A)



"""Algorithms for indexing."""

import numpy as np

from recordlinkage.measures import max_pairs


def _map_triu_1d_on_2d(indices, dims):
    """Map 1d indices on upper triangular matrix in 2d. """

    N = (dims * dims - dims) / 2

    m = np.ceil(np.sqrt(2 * N))
    c = m - np.round(np.sqrt(2 * (N - indices))) - 1
    r = np.mod(indices + (c + 1) * (c + 2) / 2 - 1, m) + 1

    return np.array([c, r], dtype=np.int64)


def _unique_rows_numpy(a):
    """return unique rows"""
    a = np.ascontiguousarray(a)
    unique_a = np.unique(a.view([('', a.dtype)] * a.shape[1]))
    return unique_a.view(a.dtype).reshape((unique_a.shape[0], a.shape[1]))


def random_pairs_with_replacement(n, shape, random_state=None):
    """make random record pairs"""

    if not isinstance(random_state, np.random.RandomState):
        random_state = np.random.RandomState(random_state)

    n_max = max_pairs(shape)

    if n_max <= 0:
        raise ValueError('n_max must be larger than 0')

    # make random pairs
    indices = random_state.randint(0, n_max, n)

    if len(shape) == 1:
        return _map_triu_1d_on_2d(indices, shape[0])
    else:
        return np.unravel_index(indices, shape)


def random_pairs_without_replacement_small_frames(
        n, shape, random_state=None):

    n_max = max_pairs(shape)

    if not isinstance(random_state, np.random.RandomState):
        random_state = np.random.RandomState(random_state)

    if not isinstance(n, int) or n <= 0 or n > n_max:
        raise ValueError("n must be a integer satisfying 0<n<=%s" % n_max)

    # make a sample without replacement
    sample = random_state.choice(
        np.arange(n_max), n, replace=False)

    # return 2d indices
    if len(shape) == 1:
        return _map_triu_1d_on_2d(sample, shape[0])
    else:
        return np.unravel_index(sample, shape)


def random_pairs_without_replacement_large_frames(
        n, shape, random_state=None):
    """Make a sample of random pairs with replacement"""

    n_max = max_pairs(shape)

    sample = np.array([])

    # Run as long as the number of pairs is less than the requested number
    # of pairs n.
    while len(sample) < n:

        # The number of pairs to sample (sample twice as much record pairs
        # because the duplicates are dropped).
        n_sample_size = (n - len(sample)) * 2
        sample = random_state.randint(n_max, size=n_sample_size)

        # concatenate pairs and deduplicate
        pairs_non_unique = np.append(sample, sample)
        sample = _unique_rows_numpy(pairs_non_unique)

    # return 2d indices
    if len(shape) == 1:
        return _map_triu_1d_on_2d(sample[0:n], shape[0])
    else:
        return np.unravel_index(sample[0:n], shape)



import pandas
import numpy as np


def _compare_exact(s1, s2, agree_value=1, disagree_value=0, missing_value=0):
    # dtypes can be hard if the passed parameters (agreement, disagreement,
    # missing_value) are of different types.
    # http://chris.friedline.net/2015-12-15-rutgers/lessons/python2/03-data-types-and-format.html

    # Convert to pandas.Series if (numpy) arrays are passed.
    if not isinstance(s1, pandas.Series):
        s1 = pandas.Series(s1, index=s1.index)

    if not isinstance(s2, pandas.Series):
        s2 = pandas.Series(s2, index=s2.index)

    # Values or agree/disagree
    if agree_value == 'value':
        compare = s1.copy()
        compare[s1 != s2] = disagree_value

    else:
        compare = pandas.Series(disagree_value, index=s1.index)
        compare[s1 == s2] = agree_value

    # Only when disagree value is not identical with the missing value
    if disagree_value != missing_value:

        compare[(s1.isnull() | s2.isnull())] = missing_value

    return compare


def _compare_dates(s1, s2, swap_month_day=0.5, swap_months='default',
                   errors='coerce', *args, **kwargs):

    # validate datatypes
    if str(s1.dtype) != 'datetime64[ns]':
        raise ValueError('Left column is not of type datetime64[ns]')

    if str(s2.dtype) != 'datetime64[ns]':
        raise ValueError('Right column is not of type datetime64[ns]')

    c = (s1 == s2).astype(np.int64)  # start with int64 (will become float64)

    # The case is which there is a swap_month_day value given.
    if (swap_month_day and swap_month_day != 0):

        c[(s1.dt.year == s2.dt.year) &
          (s1.dt.month == s2.dt.day) &
          (s1.dt.day == s2.dt.month) &
          (c != 1)] = swap_month_day

    if (swap_months and swap_months != 0):

        if swap_months == 'default':
            swap_months = [(6, 7, 0.5),
                           (7, 6, 0.5),
                           (9, 10, 0.5),
                           (10, 9, 0.5)]
        else:
            try:
                if not all([len(x) == 3 for x in swap_months]):
                    raise Exception
            except Exception:
                raise ValueError(
                    'swap_months must be a list of (first month, \
                    second month, value) tuples or lists. ')

        for month1, month2, value in swap_months:

            # if isinstance(value, float):
            #     c = c.astype(np.float64)
            # elif isinstance(value, int):
            #     c = c.astype(np.int64)
            # else:
            #     c = c.astype(object)

            c[(s1.dt.year == s2.dt.year) &
              (s1.dt.month == month1) & (s2.dt.month == month2) &
              (s1.dt.day == s2.dt.day) &
              (c != 1)] = value

    c = pandas.Series(c)
    c[s1.isnull() | s2.isnull()] = np.nan

    return c



"""Logging utilities."""

# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
#
# Modifications copyright Jonathan de Bruin 2017

# pylint: disable=unused-import

import logging as _logging
from logging import DEBUG  # noqa
from logging import ERROR  # noqa
from logging import FATAL  # noqa
from logging import INFO  # noqa
from logging import WARN  # noqa
import sys as _sys

# Determine whether we are in an interactive environment
_interactive = False
try:
    # This is only defined in interactive shells
    if _sys.ps1:
        _interactive = True
except AttributeError:
    # Even now, we may be in an interactive shell with `python -i`.
    _interactive = _sys.flags.interactive

# Scope the tensorflow logger to not conflict with users' loggers
_logger = _logging.getLogger('recordlinkage')

# If we are in an interactive environment (like jupyter), set loglevel to info
# and pipe the output to stdout
if _interactive:
    _logger.setLevel(WARN)
    _logging_target = _sys.stdout
else:
    _logging_target = _sys.stderr

# Add the output handler
_handler = _logging.StreamHandler(_logging_target)
_handler.setFormatter(_logging.Formatter(_logging.BASIC_FORMAT, None))
_logger.addHandler(_handler)

log = _logger.log
debug = _logger.debug
error = _logger.error
fatal = _logger.fatal
info = _logger.info
warn = _logger.warn
warning = _logger.warning


def get_verbosity():
    """Return how much logging output will be produced."""
    return _logger.getEffectiveLevel()


def set_verbosity(verbosity):
    """Sets the threshold for what messages will be logged."""
    _logger.setLevel(verbosity)



import pandas as pd
import numpy as np


def binary_vectors(n, n_match, m=[0.9] * 8, u=[0.1] * 8,
                   random_state=None):
    """Generate random binary comparison vectors

    This function is used to generate random comparison vectors. The result of
    each comparison is a binary value (0 or 1).

    Parameters
    ----------
    n : int
        The total number of comparison vectors.
    n_match : int
        The number of matching record pairs.
    m : list, default [0.9] * 8, optional
        A list of m probabilities of each partially identifying variable. The
        m probability is the probability that an identifier in matching
        record pairs agrees.
    u : list, default [0.9] * 8, optional
        A list of u probabilities of each partially identifying variable. The
        u probability is the probability that an identifier in non-matching
        record pairs agrees.
    random_state : int or numpy.random.RandomState, optional
        Seed for the random number generator with an integer or numpy
        RandomState object.

    Returns
    -------
    pandas.DataFrame
        A dataframe with comparison vectors.


    """

    if len(m) != len(u):
        raise ValueError("the length of 'm' is not equal the length of 'u'")

    if n_match >= n or n_match < 0:
        raise ValueError("the number of matches is bounded by [0, n]")

    # set the random seed
    np.random.seed(random_state)

    matches = []
    nonmatches = []

    for i, _ in enumerate(m):

        p_mi = [1 - m[i], m[i]]
        p_ui = [1 - u[i], u[i]]

        comp_mi = np.random.choice([0, 1], (n_match, 1), p=p_mi)
        comp_ui = np.random.choice([0, 1], (n - n_match, 1), p=p_ui)

        nonmatches.append(comp_ui)
        matches.append(comp_mi)

    match_block = np.concatenate(matches, axis=1)
    nonmatch_block = np.concatenate(nonmatches, axis=1)

    data_np = np.concatenate((match_block, nonmatch_block), axis=0)
    index_np = np.random.randint(1001, 1001 + n * 2, (n, 2))

    data_col_names = ['c_%s' % (i + 1) for i in range(len(m))]
    data_mi = pd.MultiIndex.from_arrays([index_np[:, 0], index_np[:, 1]])
    data_df = pd.DataFrame(data_np, index=data_mi, columns=data_col_names)

    return data_df.sample(frac=1, random_state=random_state)



"""Base module for record linkage."""

from __future__ import division

import time
import warnings

import pandas
import numpy as np
from sklearn.externals.joblib import Parallel, delayed

from recordlinkage.utils import listify
from recordlinkage.utils import unique
from recordlinkage.utils import is_label_dataframe
from recordlinkage.utils import VisibleDeprecationWarning
from recordlinkage.utils import split_index
from recordlinkage.types import is_pandas_like
from recordlinkage.types import is_numpy_like
from recordlinkage.measures import max_pairs
from recordlinkage import rl_logging as logging


def chunk_pandas(frame_or_series, chunksize=None):
    """Chunk a frame into smaller, equal parts."""

    if not isinstance(chunksize, int):
        raise ValueError('argument chunksize needs to be integer type')

    bins = np.arange(0, len(frame_or_series), step=chunksize)

    for b in bins:

        yield frame_or_series[b:b + chunksize]


class BaseIndexator(object):
    """Base class for all indexator classes in Python Record Linkage Toolkit.

    The structure of the indexing classes follow the framework of SciKit-
    learn and tensorflow.

    Example
    -------
    Make your own indexation class
    ```
    class CustomIndex(BaseIndexator):

        def _link_index(self, df_a, df_b):

            # Custom link index.

            return ...

        def _dedup_index(self, df_a):

            # Custom deduplication index, optional.

            return ...
    ```

    Call the class in the same way
    ```
    custom_index = CustomIndex():
    custom_index.index()
    ```
    """

    def __init__(self, verify_integrity=True):
        super(BaseIndexator, self).__init__()

        self._n = []
        self._n_max = []

        self.verify_integrity = verify_integrity

        logging.info("Indexing - initialize {} class".format(
            self.__class__.__name__)
        )

    def __repr__(self):
        class_name = self.__class__.__name__
        return "<{}>".format(class_name)

    def __str__(self):
        return repr(self)

    @classmethod
    def _deduplication(cls, x):

        if isinstance(x, (tuple, list)) and len(x) > 1:
            return False
        else:
            return True

    def _verify_integrety(self, x):

        if isinstance(x.index, pandas.Index):

            if not x.index.is_unique:
                raise ValueError('index of DataFrame is not unique')

        elif isinstance(x.index, pandas.MultiIndex):
            raise ValueError(
                'expected pandas.Index instead of pandas.MultiIndex'
            )

    def _link_index(self, df_a, df_b):

        return NotImplementedError(
            "Not possible to call index for the BaseEstimator"
        )

    def _dedup_index(self, df_a):
        """Make record pairs of a single dataframe."""

        pairs = self._link_index(df_a, df_a)

        # Remove all pairs not in the upper triangular part of the matrix.
        # This part can be inproved by not comparing the level values, but the
        # level itself.
        pairs = pairs[
            pairs.get_level_values(0) < pairs.get_level_values(1)
        ]

        return pairs

    def fit(self):

        raise AttributeError("indexing object has no attribute 'fit'")

    def index(self, x, x_link=None):
        """Make an index of record pairs.

        Use a custom function to make record pairs of one or two dataframes.
        Each function should return a pandas.MultiIndex with record pairs.

        Parameters
        ----------
        x: pandas.DataFrame
            A pandas DataFrame. When `x_link` is None, the algorithm makes
            record pairs within the DataFrame. When `x_link` is not empty,
            the algorithm makes pairs between `x` and `x_link`.
        x_link: pandas.DataFrame, optional
            A second DataFrame to link with the DataFrame x.

        Returns
        -------
        pandas.MultiIndex
            A pandas.MultiIndex with record pairs. Each record pair contains
            the index labels of two records.

        """

        if x is None:  # error
            raise ValueError("provide at least one dataframe")
        elif x_link is not None:  # linking (two arg)
            x = (x, x_link)
        elif isinstance(x, (list, tuple)):  # dedup or linking (single arg)
            x = tuple(x)
        else:  # dedup (single arg)
            x = (x,)

        if self.verify_integrity:

            for df in x:
                self._verify_integrety(df)

        # start timing
        start_time = time.time()

        # linking
        if not self._deduplication(x):

            logging.info("Indexing - start indexing two DataFrames")

            pairs = self._link_index(*x)

        # deduplication
        else:
            logging.info("Indexing - start indexing single DataFrame")

            pairs = self._dedup_index(*x)

        # store the number of pairs
        self._n.append(pairs.shape[0])
        self._n_max.append(max_pairs(x))

        # summary
        n = len(pairs)
        rr = 1 - self._n[-1] / self._n_max[-1]
        rr_avg = 1 - np.sum(self._n) / np.sum(self._n_max)

        # log timing
        logf_time = "Indexing - computation time: ~{:.2f}s"
        logging.info(logf_time.format(time.time() - start_time))

        # log results
        logf_result = "Indexing - summary n={:d}, " \
            "reduction_ratio={:0.5f}, reduction_ratio_mean={:0.5f}"
        logging.info(logf_result.format(n, rr, rr_avg))

        return pairs


class CompareFeature(object):
    """CompareFeature construction class.
    """

    name = None

    def __init__(self, comp_func, labels_left, labels_right, args=(),
                 kwargs={}, label=None, name="", description=""):

        self.comp_func = comp_func
        self.labels_left = labels_left
        self.labels_right = labels_right
        self.args = args
        self.kwargs = kwargs
        self.label = label
        self.description = description

        # logging
        logging.info(
            "CompareFeature - initialize exact algorithm - compare "
            "{l_left} with {l_right}".format(l_left=labels_left,
                                             l_right=labels_right)
        )

    def __repr__(self):

        return "<CompareFeature {!r}>".format(self.label)

    def __str__(self):
        return repr(self)

def _compute_parallel(compare_obj, pairs, x, x_link=None, n_jobs=1):

    df_chunks = split_index(pairs, n_jobs)
    result_chunks = Parallel(n_jobs=n_jobs)(
        delayed(_compute)(compare_obj, chunk, x, x_link) for chunk in df_chunks
    )

    result = pandas.concat(result_chunks)
    return result


def _compute(self, pairs, x, x_link=None):

    logging.info("Comparing - start comparing data")

    # start the timer for the comparing step
    start_time = time.time()

    sublabels_left = self._get_labels_left(validate=x)
    df_a_indexed = self._loc2(x[sublabels_left], pairs, 0)

    if x_link is None:
        sublabels_right = self._get_labels_right(validate=x)
        df_b_indexed = self._loc2(x[sublabels_right], pairs, 1)
    else:
        sublabels_right = self._get_labels_right(validate=x_link)
        df_b_indexed = self._loc2(x_link[sublabels_right], pairs, 1)

    # log timing
    index_time = time.time() - start_time

    results = pandas.DataFrame(index=pairs)
    label_num = 0  # make a label is label is None

    for feat in self._compare_functions:

        lbl1 = feat.labels_left
        lbl2 = feat.labels_right
        f = feat.comp_func
        label = feat.label

        data1 = tuple([df_a_indexed[lbl] for lbl in listify(lbl1)])
        data2 = tuple([df_b_indexed[lbl] for lbl in listify(lbl2)])

        c = f(*tuple(data1 + data2 + feat.args), **feat.kwargs)

        if isinstance(c, (pandas.Series, pandas.DataFrame)):
            c = c.values  # convert pandas into numpy

        if label is not None:
            label = listify(label)

        n_cols = 1 if len(c.shape) == 1 else c.shape[1]

        labels = []
        for i in range(0, n_cols):

            label_val = label[i] if label is not None else label_num
            label_num += 1

            labels.append(label_val)

        results[label_val] = c

    # log timing
    total_time = time.time() - start_time

    # log timing
    logging.info("Comparing - computation time: ~{:.2f}s (from which "
                 "indexing: ~{:.2f}s)".format(total_time, index_time))

    # log results
    logf_result = "Comparing - summary shape={}"
    logging.info(logf_result.format(results.shape))

    return results


class BaseCompare(object):
    """Base class for all comparing classes in Python Record Linkage Toolkit.
    """

    def __init__(self, pairs=None, df_a=None, df_b=None, low_memory=False,
                 block_size=1000000, n_jobs=1, indexing_type='label',
                 **kwargs):

        logging.info("Comparing - initialize {} class".format(
            self.__class__.__name__)
        )

        # public
        self.indexing_type = indexing_type  # label of position

        # private
        self._compare_functions = []

        if isinstance(pairs, (pandas.MultiIndex, pandas.Index)):
            self.deprecated = True

            warnings.warn(
                "It seems you are using the older version of the Compare API, "
                "see the documentation about how to update to the new API. "
                "http://recordlinkage.readthedocs.io/"
                "en/latest/ref-compare.html",
                VisibleDeprecationWarning
            )
        else:
            self.deprecated = False

        # start deprecated
        self.df_a = df_a
        self.df_b = df_b if df_b is not None else df_a

        self.pairs = pairs

        self.low_memory = low_memory
        self.block_size = block_size
        self.n_jobs = n_jobs

        self._df_a_indexed = None
        self._df_b_indexed = None

        self.vectors = pandas.DataFrame(index=pairs)
        # end deprecated

    def __repr__(self):
        class_name = self.__class__.__name__
        return "<{}>".format(class_name)

    def __str__(self):
        return repr(self)

    def _loc2(self, frame, multi_index, level_i, indexing_type='label'):
        """Indexing algorithm for MultiIndex on one level

        Arguments
        ---------
        frame : pandas.DataFrame
            The datafrme to select records from.
        multi_index : pandas.MultiIndex
            A pandas multiindex were one fo the levels is used to sample the
            dataframe with.
        level_i : int, str
            The level of the multiIndex to index on.
        indexing_type : str
            The type of indexing. The value can be 'label' or 'position'.
            Default 'label'.

        """

        if indexing_type == "label":
            data = frame.loc[multi_index.get_level_values(level_i)]
            data.index = multi_index
        elif indexing_type == "position":
            data = frame.iloc[multi_index.get_level_values(level_i)]
            data.index = multi_index
        else:
            raise ValueError("indexing_type needs to be 'label' or 'position'")

        return data

    def compare_vectorized(self, comp_func, labels_left, labels_right,
                           *args, **kwargs):
        """Compute the similarity between values with a callable.

        This method initialises the comparing of values with a custom
        function/callable. The function/callable should accept
        numpy.ndarray's.

        Example
        -------

        >>> comp = recordlinkage.Compare()
        >>> comp.compare_vectorized(custom_callable, 'first_name', 'name')
        >>> comp.compare(PAIRS, DATAFRAME1, DATAFRAME2)

        Parameters
        ----------
        comp_func : function
            A comparison function. This function can be a built-in function
            or a user defined comparison function. The function should accept
            numpy.ndarray's as first two arguments.
        labels_left : label, pandas.Series, pandas.DataFrame
            The labels, Series or DataFrame to compare.
        labels_right : label, pandas.Series, pandas.DataFrame
            The labels, Series or DataFrame to compare.
        *args :
            Additional arguments to pass to callable comp_func.
        **kwargs :
            Additional keyword arguments to pass to callable comp_func.
            (keyword 'label' is reserved.)
        label : (list of) label(s)
            The name of the feature and the name of the column. IMPORTANT:
            This argument is a keyword argument and can not be part of the
            arguments of comp_func.
        """

        label = kwargs.pop('label', None)

        return self._compare_vectorized(
            comp_func, labels_left, labels_right, args, kwargs, label=label)

    def _compare_vectorized(self, comp_func, labels_left, labels_right,
                            args=(), kwargs={}, label=None, name="",
                            description=""):
        # This function is separated from the compare_vectorized method to
        # solve conficts between the arguments of comp_func and
        # _compare_vectorized. Only the keyword 'label' is preserved for
        # compare_vectorized, but not 'name' and 'description'. The args and
        # kwargs are normal (keyword) arguments of _compare_vectorized (so not
        # *args and **kwargs). Therefore, the compare_vectorized is a slightly
        # more limit version because only 'label' is an preserved keyword. For
        # advanced usage, one can directly call _compare_vectorized instead of
        # comapre_vectorized. This is done in the methods of the Compare
        # class.

        # Use recordlinkage >=0.10.0
        if not self.deprecated:

            if isinstance(labels_left, tuple):
                labels_left = list(labels_left)

            if isinstance(labels_right, tuple):
                labels_right = list(labels_right)

            feat = CompareFeature(comp_func, labels_left, labels_right,
                                  args, kwargs, label=label, name=name,
                                  description=description)

            self._compare_functions.append(feat)

            # return self to make pipelining possible
            return self

        # Use recordlinkage < 0.10.0
        else:

            return self.compare(
                comp_func, labels_left, labels_right, *args, **kwargs)

    def _get_labels_left(self, validate=None):
        """Get all labels of the left dataframe."""

        labels = []

        for compare_func in self._compare_functions:

            labels = labels + listify(compare_func.labels_left)

        # check requested labels (for better error messages)
        if not is_label_dataframe(labels, validate):
            error_msg = "label is not found in the dataframe"
            raise KeyError(error_msg)

        return unique(labels)

    def _get_labels_right(self, validate=None):
        """Get all labels of the right dataframe."""
        labels = []

        for compare_func in self._compare_functions:

            labels = labels + listify(compare_func.labels_right)

        # check requested labels (for better error messages)
        if not is_label_dataframe(labels, validate):
            error_msg = "label is not found in the dataframe"
            raise KeyError(error_msg)

        return unique(labels)

    def compute(self, pairs, x, x_link=None):
        """Compare the records of each record pair.

        Calling this method starts the comparing of records.

        Parameters
        ----------
        pairs : pandas.MultiIndex
            A pandas MultiIndex with the record pairs to compare. The indices
            in the MultiIndex are indices of the DataFrame(s) to link.
        x : pandas.DataFrame
            The DataFrame to link. If `x_link` is given, the comparing is a
            linking problem. If `x_link` is not given, the problem is one of
            deduplication.
        x_link : pandas.DataFrame, optional
            The second DataFrame.

        Returns
        -------
        pandas.DataFrame
            A pandas DataFrame with feature vectors, i.e. the result of
            comparing each record pair.
        """

        if not isinstance(pairs, pandas.MultiIndex):
            raise ValueError(
                "expected pandas.MultiIndex with record pair indices "
                "as first argument"
            )

        if not isinstance(x, pandas.DataFrame):
            raise ValueError("expected pandas.DataFrame as second argument")

        if x_link is not None and not isinstance(x_link, pandas.DataFrame):
            raise ValueError("expected pandas.DataFrame as third argument")

        if self.n_jobs == 1:
            results = _compute(self, pairs, x, x_link)
        elif self.n_jobs > 1:
            results = _compute_parallel(self, pairs, x, x_link,
                                        n_jobs=self.n_jobs)
        else:
            raise ValueError("number of jobs should be positive integer")

        return results

    def _compute_parallel(self, pairs, x, x_link):

        return _compute_parallel(self, pairs, x, x_link, self.n_jobs)

    def _compute(self, pairs, x, x_link):

        return _compute(self, pairs, x, x_link)

    def compare(self, comp_func, labels_a, labels_b, *args, **kwargs):
        """[DEPRECATED] Compare two records.

        Core method to compare record pairs. This method takes a function and
        data from both records in the record pair. The data is compared with
        the compare function. The built-in methods also use this function.

        Example
        -------

        >>> comp = recordlinkage.Compare(PAIRS, DATAFRAME1, DATAFRAME2)
        >>> comp.exact('first_name', 'name')

        >>> # same as
        >>> comp.compare(recordlinkage._compare_exact, 'first_name', 'name')

        Parameters
        ----------
        comp_func : function
            A comparison function. This function can be a built-in function or
            a user defined comparison function.
        labels_a : label, pandas.Series, pandas.DataFrame
            The labels, Series or DataFrame to compare.
        labels_b : label, pandas.Series, pandas.DataFrame
            The labels, Series or DataFrame to compare.
        name : label
            The name of the feature and the name of the column.
        store : bool, default True
            Store the result in the dataframe.

        Returns
        -------
        pandas.Series
            A pandas series with the result of comparing each record pair.

        """

        if isinstance(comp_func, pandas.MultiIndex):
            raise ValueError("see new api documentation: "
                             "use method 'compute' instead of 'compare'")

        if len(self.pairs) == 0:
            raise ValueError(
                "need at least one record pair"
            )

        # the name and store arguments
        name = kwargs.pop('name', None)
        store = kwargs.pop('store', True)

        labels_a = listify(labels_a)
        labels_b = listify(labels_b)

        data_a = []

        for label_a in labels_a:

            # the label is a numpy or pandas object
            if is_numpy_like(label_a) or is_pandas_like(label_a):
                data_a.append(label_a)

            # check requested labels (for better error messages)
            elif label_a not in self.df_a.columns:
                raise KeyError("label '{}' is not found in the first"
                               "dataframe".format(label_a))

            else:

                if self.low_memory:

                    df_a_label = self._loc2(self.df_a[label_a], self.pairs, 0)
                    data_a.append(df_a_label)

                # not low memory
                else:
                    if self._df_a_indexed is None:

                        self._df_a_indexed = self._loc2(
                            self.df_a, self.pairs, 0)

                    data_a.append(self._df_a_indexed[label_a])

        data_a = tuple(data_a)

        data_b = []

        for label_b in labels_b:

            # the label is a numpy or pandas object
            if is_numpy_like(label_b) or is_pandas_like(label_b):
                data_b.append(label_b)

            # check requested labels (for better error messages)
            elif label_b not in self.df_b.columns:

                raise KeyError("label '{}' is not found in the second"
                               "dataframe".format(label_b))

            else:

                if self.low_memory:

                    df_b_label = self._loc2(self.df_b[label_b], self.pairs, 1)
                    data_b.append(df_b_label)

                # not low memory
                else:
                    if self._df_b_indexed is None:

                        self._df_b_indexed = self._loc2(
                            self.df_b, self.pairs, 1)

                    data_b.append(self._df_b_indexed[label_b])

        data_b = tuple(data_b)

        # Compute the comparison
        c = comp_func(*tuple(data_a + data_b + args), **kwargs)

        # if a pandas series is returned, overwrite the index. The
        # returned index can be different than the MultiIndex passed to
        # the compare function.
        if isinstance(c, pandas.Series):
            c.index = self.vectors.index

        # append column to Compare.vectors
        if store:
            name_or_id = name if name else len(self.vectors.columns)
            self.vectors[name_or_id] = c

        return self.vectors[name_or_id].rename(name)

    def clear_memory(self):
        """[DEPRECATED] Clear memory.

        Clear some memory when low_memory was set to True.
        """

        self._df_a_indexed = None
        self._df_b_indexed = None



import sys

import numpy as np
import pandas

import jellyfish


def phonetic(s, method, concat=True, encoding='utf-8', decode_error='strict'):
    """Convert names or strings into phonetic codes. 

    The implemented algorithms are `soundex
    <https://en.wikipedia.org/wiki/Soundex>`_, `nysiis
    <https://en.wikipedia.org/wiki/New_York_State_Identification_and_
    Intelligence_System>`_, `metaphone
    <https://en.wikipedia.org/wiki/Metaphone>`_ or  `match_rating
    <https://en.wikipedia.org/wiki/Match_rating_approach>`_.

    Parameters
    ----------
    method: string
        The algorithm that is used to phonetically encode the values. The
        possible options are "soundex", "nysiis", "metaphone" or
        "match rating".
    concat: bool, optional
        Remove whitespace before phonetic encoding.
    encoding: string, optional
        If bytes are given, this encoding is used to decode. Default is
        'utf-8'.
    decode_error: {'strict', 'ignore', 'replace'}, optional
        Instruction on what to do if a byte Series is given that contains
        characters not of the given `encoding`. By default, it is 'strict',
        meaning that a UnicodeDecodeError will be raised. Other values are
        'ignore' and 'replace'.:type method: str

    Returns
    -------
    pandas.Series
        A Series with phonetic encoded values.

    Note
    ----
    The 'soundex' and 'nysiis' algorithms use the package 'jellyfish'. It can
    be installed with pip (``pip install jellyfish``).

    """

    # encoding
    if sys.version_info[0] == 2:
        s = s.apply(
            lambda x: x.decode(encoding, decode_error) if type(x) == bytes else x)

    if concat:
        s = s.str.replace('[\-\_\s]', '')

    if method == 'soundex':
        phonetic_callback = jellyfish.soundex
    elif method == 'nysiis':
        phonetic_callback = jellyfish.nysiis
    elif method == 'metaphone':
        phonetic_callback = jellyfish.metaphone
    elif method in ['match_rating', 'match rating', 'matchrating', 'match_rating_codex', 'matchratingcodex']:
        phonetic_callback = jellyfish.match_rating_codex
    else:
        raise ValueError("The algorithm '{}' is not known.".format(method))

    return s.str.upper().apply(
        lambda x: phonetic_callback(x) if pandas.notnull(x) else np.nan
    )



from __future__ import division
from __future__ import absolute_import
# from __future__ import unicode_literals

from sklearn.feature_extraction.text import strip_accents_ascii, \
    strip_accents_unicode


def clean(s, lowercase=True, replace_by_none=r'[^ \-\_A-Za-z0-9]+',
          replace_by_whitespace=r'[\-\_]', strip_accents=None,
          remove_brackets=True, encoding='utf-8', decode_error='strict'):
    """Clean string variables.

    Clean strings in the Series by removing unwanted tokens, whitespace and
    brackets.

    Parameters
    ----------
    s : pandas.Series
        A Series to clean.
    lower : bool, optional
        Convert strings in the Series to lowercase. Default True.
    replace_by_none : str, optional
        The matches of this regular expression are replaced by ''.
    replace_by_whitespace : str, optional
        The matches of this regular expression are replaced by a whitespace.
    remove_brackets : bool, optional
        Remove all content between brackets and the brackets themselves.
        Default True.
    strip_accents : {'ascii', 'unicode', None}, optional
        Remove accents during the preprocessing step. 'ascii' is a fast method
        that only works on characters that have an direct ASCII mapping.
        'unicode' is a slightly slower method that works on any characters.
        None (default) does nothing.
    encoding : string, optional
        If bytes are given, this encoding is used to decode. Default is
        'utf-8'.
    decode_error : {'strict', 'ignore', 'replace'}, optional
        Instruction on what to do if a byte Series is given that contains
        characters not of the given `encoding`. By default, it is 'strict',
        meaning that a UnicodeDecodeError will be raised. Other values are
        'ignore' and 'replace'.

    Example
    -------
    >>> import pandas
    >>> from recordlinkage.preprocessing import clean
    >>>
    >>> name = ['Mary-ann', 'Bob :)', 'Angel', 'Bob (alias Billy)', None]
    >>> s = pandas.Series(names)
    >>> print(clean(s))
    0    mary ann
    1         bob
    2       angel
    3         bob
    4         NaN
    dtype: object

    Returns
    -------
    pandas.Series:
        A cleaned Series of strings.

    """

    if s.shape[0] == 0:
        return s

    # Lower s if lower is True
    if lowercase is True:
        s = s.str.lower()

    # Accent stripping based on https://github.com/scikit-learn/
    # scikit-learn/blob/412996f/sklearn/feature_extraction/text.py
    # BSD license
    if not strip_accents:
        pass
    elif callable(strip_accents):
        strip_accents_fn = strip_accents
    elif strip_accents == 'ascii':
        strip_accents_fn = strip_accents_ascii
    elif strip_accents == 'unicode':
        strip_accents_fn = strip_accents_unicode
    else:
        raise ValueError(
            "Invalid value for 'strip_accents': {}".format(strip_accents)
        )

    # Remove accents etc
    if strip_accents:

        # encoding
        s = s.apply(
            lambda x: x.decode(encoding, decode_error) if type(x) == bytes else x)
        s = s.map(lambda x: strip_accents_fn(x))

    # Remove all content between brackets
    if remove_brackets is True:
        s = s.str.replace(r'(\[.*?\]|\(.*?\)|\{.*?\})', '')

    # Remove the special characters
    if replace_by_none:
        s = s.str.replace(replace_by_none, '')

    if replace_by_whitespace:
        s = s.str.replace(replace_by_whitespace, ' ')

    # Remove multiple whitespaces
    s = s.str.replace(r'\s\s+', ' ')

    # Strip s
    s = s.str.lstrip().str.rstrip()

    return s


def phonenumbers(s):
    """Clean phonenumbers by removing all non-numbers (except +).

    Parameters
    ----------
    s: pandas.Series
        A Series to clean.

    Returns
    -------
    pandas.Series
        A Series with cleaned phonenumbers.

    """

    # Remove all special tokens
    s = s.astype(object).str.replace('[^0-9+]+', '')

    return s


def value_occurence(s):
    """Count the number of times each value occurs.

    This function returns the counts for each row, in contrast with
    `pandas.value_counts <http://pandas.pydata.org/pandas-
    docs/stable/generated/pandas.Series.value_counts.html>`_.

    Returns
    -------
    pandas.Series
        A Series with value counts.

    """

    # https://github.com/pydata/pandas/issues/3729
    value_count = s.fillna('NAN')

    return value_count.groupby(by=value_count).transform('count')



import pandas
import numpy as np


# Numerical distance algorithms
def _1d_distance(s1, s2):

    return pandas.eval("s2-s1")


def _haversine_distance(lat1, lng1, lat2, lng2):

    # degrees to radians conversion
    to_rad = np.deg2rad(1)

    # numeric expression to use with numexpr package
    expr = '2*6371*arcsin(sqrt((sin((lat2*to_rad-lat1*to_rad)/2))**2+cos(lat1*to_rad)*cos(lat2*to_rad)*(sin((lng2*to_rad-lng1*to_rad)/2))**2))'

    return pandas.eval(expr)



import os
import zipfile

import pandas

from six import BytesIO
from six.moves.urllib.request import urlopen


def load_krebsregister(block=[1, 2, 3, 4, 5, 6, 7, 8, 9, 10], missing_values=None, shuffle=True):
    """Dataset 'Krebsregister'

    This dataset of comparison patterns was obtained in a epidemiological
    cancer study in Germany. The comparison patterns were created by the
    Institute for Medical Biostatistics, Epidemiology and Informatics (IMBEI)
    and the University Medical Center of Johannes Gutenberg University
    (Mainz, Germany). The dataset is available for research online.

        "The records represent individual data including first and family
        name, sex, date of birth and postal code, which were collected
        through iterative insertions in the course of several years. The
        comparison patterns in this data set are based on a sample of
        100.000 records dating from 2005 to 2008. Data pairs were
        classified as "match" or "non-match" during  an extensive manual
        review where several documentarists were involved.  The resulting
        classification formed the basis for assessing the quality of the
        registry's own record linkage procedure.

        In order to limit the amount of patterns a blocking procedure was
        applied, which selects only record pairs that meet specific
        agreement conditions. The results of the following six blocking
        iterations were merged together:

        1. Phonetic equality of first name and family name, equality of date of birth.
        2. Phonetic equality of first name, equality of day of birth.
        3. Phonetic equality of first name, equality of month of birth.
        4. Phonetic equality of first name, equality of year of birth.
        5. Equality of complete date of birth.
        6. Phonetic equality of family name, equality of sex.

        This procedure resulted in 5.749.132 record pairs, of which
        20.931 are matches. The data set is split into 10 blocks of
        (approximately) equal size and ratio of matches to non-matches."

    Parameters
    ----------
    block : int, list
        An integer or a list with integers between 1 and 10. The blocks are
        the blocks explained in the description.
    missing_values : object, int, float
        The value of the missing values. Default NaN.
    shuffle : bool
        Shuffle the record pairs. Default True.

    Returns
    -------
    (pandas.DataFrame, pandas.MultiIndex)
        A data frame with comparison vectors and a multi index with the
        indices of the matches.

    """

    # If the data is not found, download it.
    for i in range(1, 11):

        filepath = os.path.join(os.path.dirname(__file__),
                                'krebsregister', 'block_{}.zip'.format(i))

        if not os.path.exists(filepath):
            _download_krebsregister()
            break

    if isinstance(block, (list, tuple)):

        data = pandas.concat([_krebsregister_block(bl) for bl in block])
    else:

        data = _krebsregister_block(block)

    if shuffle:
        data = data.sample(frac=1, random_state=535)

    match_index = data.index[data['is_match']]
    del data['is_match']

    if pandas.notnull(missing_values):
        data.fillna(missing_values, inplace=True)

    return data, match_index


def _download_krebsregister():

    zip_file_url = "http://archive.ics.uci.edu/ml/machine-learning-databases/00210/donation.zip"

    try:
        print("Start downloading the data.")
        r = urlopen(zip_file_url).read()

        # unzip the content and put it in the krebsregister folder
        z = zipfile.ZipFile(BytesIO(r))
        z.extractall(os.path.join(os.path.dirname(__file__), 'krebsregister'))

        print("Data download succesfull.")

    except Exception as e:
        print("Issue with downloading the data:", e)


def _krebsregister_block(block):

    if block not in [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]:
        raise ValueError(
            "Argument 'block' has to be integer in [1, 2, 3, 4, 5, 6, 7, 8, 9, 10] or list of integers.")

    fp_i = os.path.join(os.path.dirname(__file__),
                        'krebsregister', 'block_{}.zip'.format(block))

    data_block = pandas.read_csv(
        fp_i,
        index_col=['id_1', 'id_2'],
        na_values='?',
        compression='zip')

    data_block.columns = [
        'cmp_firstname1', 'cmp_firstname2', 'cmp_lastname1', 'cmp_lastname2',
        'cmp_sex', 'cmp_birthday', 'cmp_birthmonth', 'cmp_birthyear',
        'cmp_zipcode', 'is_match']
    data_block.index.names = ['id1', 'id2']

    return data_block



from __future__ import division
from __future__ import unicode_literals

import warnings

import pandas
import numpy as np
import jellyfish

from sklearn.feature_extraction.text import CountVectorizer


# Ingore zero devision errors in cosine and qgram algorithms
# warnings.filterwarnings("ignore")

################################
#      STRING SIMILARITY       #
################################


def jaro_similarity(s1, s2):

    conc = pandas.Series(list(zip(s1, s2)))

    def jaro_apply(x):

        try:
            return jellyfish.jaro_distance(x[0], x[1])
        except Exception as err:
            if pandas.isnull(x[0]) or pandas.isnull(x[1]):
                return np.nan
            else:
                raise err

    return conc.apply(jaro_apply)


def jarowinkler_similarity(s1, s2):

    conc = pandas.Series(list(zip(s1, s2)))

    def jaro_winkler_apply(x):
        try:
            return jellyfish.jaro_winkler(x[0], x[1])
        except Exception as err:
            if pandas.isnull(x[0]) or pandas.isnull(x[1]):
                return np.nan
            else:
                raise err

    return conc.apply(jaro_winkler_apply)


def levenshtein_similarity(s1, s2):

    conc = pandas.Series(list(zip(s1, s2)))

    def levenshtein_apply(x):

        try:
            return 1 - jellyfish.levenshtein_distance(x[0], x[1]) \
                / np.max([len(x[0]), len(x[1])])
        except Exception as err:
            if pandas.isnull(x[0]) or pandas.isnull(x[1]):
                return np.nan
            else:
                raise err

    return conc.apply(levenshtein_apply)


def damerau_levenshtein_similarity(s1, s2):

    conc = pandas.Series(list(zip(s1, s2)))

    def damerau_levenshtein_apply(x):

        try:
            return 1 - jellyfish.damerau_levenshtein_distance(x[0], x[1]) \
                / np.max([len(x[0]), len(x[1])])
        except Exception as err:
            if pandas.isnull(x[0]) or pandas.isnull(x[1]):
                return np.nan
            else:
                raise err

    return conc.apply(damerau_levenshtein_apply)


def qgram_similarity(s1, s2, include_wb=True, ngram=(2, 2)):

    if len(s1) != len(s2):
        raise ValueError('Arrays or Series have to be same length.')

    if len(s1) == len(s2) == 0:
        return []

    # include word boundaries or not
    analyzer = 'char_wb' if include_wb is True else 'char'

    # prepare data
    data = s1.append(s2).fillna('')

    # The vectorizer
    vectorizer = CountVectorizer(
        analyzer=analyzer, 
        strip_accents='unicode', 
        ngram_range=ngram
    )

    vec_fit = vectorizer.fit_transform(data)

    def _metric_sparse_euclidean(u, v):

        match_ngrams = u.minimum(v).sum(axis=1)
        total_ngrams = np.maximum(u.sum(axis=1), v.sum(axis=1))

        # division by zero is not possible in our case, but 0/0 is possible.
        # Numpy raises a warning in that case.
        return np.true_divide(match_ngrams, total_ngrams).A1

    return _metric_sparse_euclidean(vec_fit[:len(s1)], vec_fit[len(s1):])


def cosine_similarity(s1, s2, include_wb=True, ngram=(2, 2)):

    if len(s1) != len(s2):
        raise ValueError('Arrays or Series have to be same length.')

    if len(s1) == len(s2) == 0:
        return []

    # include word boundaries or not
    analyzer = 'char_wb' if include_wb is True else 'char'

    # The vectorizer
    vectorizer = CountVectorizer(
        analyzer=analyzer, strip_accents='unicode', ngram_range=ngram)

    data = s1.append(s2).fillna('')

    vec_fit = vectorizer.fit_transform(data)

    def _metric_sparse_cosine(u, v):

        a = np.sqrt(u.multiply(u).sum(axis=1))
        b = np.sqrt(v.multiply(v).sum(axis=1))

        ab = v.multiply(u).sum(axis=1)

        return np.divide(ab, np.multiply(a, b)).A1

    return _metric_sparse_cosine(vec_fit[:len(s1)], vec_fit[len(s1):])


def smith_waterman_similarity(s1, s2, match=5, mismatch=-5, gap_start=-5, gap_continue=-1, norm="mean"):
    """
    smith_waterman_similarity(s1, s2, match=1, mismatch=-1, gap_start=-1, gap_continue=-0.2, norm="mean")

    An implementation of the Smith-Waterman string comparison algorithm
    described in Christen, Peter (2012).

    Parameters
    ----------
    s1 : label, pandas.Series
        Series or DataFrame to compare all fields.
    s2 : label, pandas.Series
        Series or DataFrame to compare all fields.
    match : float
        The value added to the match score if two characters match.
        Greater than mismatch, gap_start, and gap_continue. Default: 1.
    mismatch : float
        The value added to the match score if two characters do not match.
        Less than match. Default: -1.
    gap_start : float
        The value added to the match score upon encountering the start of
        a gap. Default: -1.
    gap_continue : float
        The value added to the match score for positions where a previously
        started gap is continuing. Default: -0.2.
    norm : str
        The name of the normalization metric to be used. Applied by dividing
        the match score by the normalization metric multiplied by match. One
        of "min", "max",or "mean". "min" will use the minimum string length
        as the normalization metric. "max" and "mean" use the maximum and
        mean string length respectively. Default: "mean""

    Returns
    -------
    pandas.Series
        A pandas series with similarity values. Values equal or between 0
        and 1.
    """
    # Assert that match is greater than or equal to mismatch, gap_start, and gap_continue.
    assert match >= max(mismatch, gap_start, gap_continue), \
        "match must be greater than or equal to mismatch, gap_start, and gap_continue"

    if len(s1) != len(s2):
        raise ValueError('Arrays or Series have to be same length.')

    if len(s1) == len(s2) == 0:
        return []

    concat = pandas.Series(list(zip(s1, s2)))

    def sw_apply(t):
        """
        sw_apply(t)

        A helper function that is applied to each pair of records
        in s1 and s2. Assigns a similarity score to each pair,
        between 0 and 1. Used by the pandas.apply method.

        Parameters
        ----------
        t : pandas.Series
            A pandas Series containing two strings to be compared.

        Returns
        -------
        Float
            A similarity score between 0 and 1.
        """
        str1 = t[0]
        str2 = t[1]

        def compute_score():
            """
            compute_score()

            The helper function that produces the non-normalized
            similarity score between two strings. The scores are
            determined using the Smith-Waterman dynamic programming
            algorithm. The scoring scheme is determined from the
            parameters provided to the parent smith_waterman_similarity()
            function.

            Returns
            -------
            Float
                A score 0 or greater. Indicates similarity between two strings.
            """

            # Initialize the score matrix with 0s

            m = [[0] * (1 + len(str2)) for i in range(1 + len(str1))]

            # Initialize the trace matrix with empty lists
            trace = [[[] for _ in range(1 + len(str2))] for _ in range(1 + len(str1))]

            # Initialize the highest seen score to 0
            highest = 0

            # Iterate through the matrix
            for x in range(1, 1 + len(str1)):
                for y in range(1, 1 + len(str2)):
                    # Calculate Diagonal Score
                    if str1[x-1] == str2[y-1]:
                        # If characters match, add the match score to the diagonal score
                        diagonal = m[x-1][y-1] + match
                    else:
                        # If characters do not match, add the mismatch score to the diagonal score
                        diagonal = m[x-1][y-1] + mismatch

                    # Calculate the Left Gap Score
                    if "H" in trace[x-1][y]:
                        # If cell to the left's score was calculated based on a horizontal gap,
                        # add the gap continuation penalty to the left score.
                        gap_horizontal = m[x-1][y] + gap_continue
                    else:
                        # Otherwise, add the gap start penalty to the left score
                        gap_horizontal = m[x-1][y] + gap_start

                    # Calculate the Above Gap Score
                    if "V" in trace[x][y-1]:
                        # If above cell's score was calculated based on a vertical gap,
                        # add the gap continuation penalty to the above score.
                        gap_vertical = m[x][y-1] + gap_continue
                    else:
                        # Otherwise, add the gap start penalty to the above score
                        gap_vertical = m[x][y-1] + gap_start

                    # Choose the highest of the three scores
                    score = max(diagonal, gap_horizontal, gap_vertical)

                    if score <= 0:
                        # If score is less than 0, boost to 0
                        score = 0
                    else:
                        # If score is greater than 0, determine whether it was
                        # calculated based on a diagonal score, horizontal gap,
                        # or vertical gap. Store D, H, or V in the trace matrix
                        # accordingly.
                        if score == diagonal:
                            trace[x][y].append("D")
                        if score == gap_horizontal:
                            trace[x][y].append("H")
                        if score == gap_vertical:
                            trace[x][y].append("V")

                    # If the cell's score is greater than the highest score
                    # previously present, record the score as the highest.
                    if score > highest:
                        highest = score

                    # Set the cell's score to score
                    m[x][y] = score

            # After iterating through the entire matrix, return the highest
            # score found.
            return highest

        def normalize(score):
            """
            normalize(score)

            A helper function used to normalize the score produced by
            compute_score() to a score between 0 and 1. The method for
            normalization is determined by the norm argument provided
            to the parent, smith_waterman_similarity function.

            Parameters
            ----------
            score : Float
                The score produced by the compute_score() function.

            Returns
            -------
            Float
                A normalized score between 0 and 1.
            """
            if norm == "min":
                # Normalize by the shorter string's length
                return score/(min(len(str1), len(str2)) * match)
            if norm == "max":
                # Normalize by the longer string's length
                return score/(max(len(str1), len(str2)) * match)
            if norm == "mean":
                # Normalize by the mean length of the two strings
                return 2*score/((len(str1) + len(str2)) * match)
            else:
                warnings.warn('Unrecognized longest common substring normalization. Defaulting to "mean" method.')
                return 2*score/((len(str1) + len(str2)) * match)
        try:
            if len(str1) == 0 or len(str2) == 0:
                return 0
            return normalize(compute_score())

        except Exception as err:
            if pandas.isnull(t[0]) or pandas.isnull(t[1]):
                return np.nan
            else:
                raise err

    return concat.apply(sw_apply)


def longest_common_substring_similarity(s1, s2, norm='dice', min_len=2):
    """
    longest_common_substring_similarity(s1, s2, norm='dice', min_len=2)

    An implementation of the longest common substring similarity algorithm
    described in Christen, Peter (2012).

    Parameters
    ----------
    s1 : label, pandas.Series
        Series or DataFrame to compare all fields.
    s2 : label, pandas.Series
        Series or DataFrame to compare all fields.
    norm : str
        The name of the normalization applied to the raw length computed by
        the lcs algorithm. One of "overlap", "jaccard", or "dice". Default:
        "dice""

    Returns
    -------
    pandas.Series
        A pandas series with normalized similarity values.
    """

    if len(s1) != len(s2):
        raise ValueError('Arrays or Series have to be same length.')

    if len(s1) == len(s2) == 0:
        return []

    conc = pandas.Series(list(zip(s1, s2)))

    def lcs_iteration(x):
        """
        lcs_iteration(x)

        A helper function implementation of a single iteration longest common substring algorithm,
        adapted from
        https://en.wikibooks.org/wiki/Algorithm_Implementation/Strings/Longest_common_substring.
        but oriented towards the iterative approach described by Christen, Peter (2012).

        Parameters
        ----------
        x : A series containing the two strings to be compared.

        Returns
        -------
        A tuple of strings and a substring length i.e. ((str, str), int).
        """

        str1 = x[0]
        str2 = x[1]

        if str1 is np.nan or str2 is np.nan or min(len(str1), len(str2)) < min_len:
            longest = 0
            new_str1 = None
            new_str2 = None
        else:
            # Creating a matrix of 0s for preprocessing
            m = [[0] * (1 + len(str2)) for _ in range(1 + len(str1))]

            # Track length of longest substring seen
            longest = 0

            # Track the ending position of this substring in str1 (x) and str2(y)
            x_longest = 0
            y_longest = 0

            # Create matrix of substring lengths
            for x in range(1, 1 + len(str1)):
                for y in range(1, 1 + len(str2)):
                    # Check if the chars match
                    if str1[x - 1] == str2[y - 1]:
                        # add 1 to the diagonal
                        m[x][y] = m[x - 1][y - 1] + 1
                        # Update values if longer than previous longest substring
                        if m[x][y] > longest:
                            longest = m[x][y]
                            x_longest = x
                            y_longest = y
                    else:
                        # If there is no match, start from zero
                        m[x][y] = 0

            # Copy str1 and str2, but subtract the longest common substring
            # for the next iteration.
            new_str1 = str1[0:x_longest-longest]+str1[x_longest:]
            new_str2 = str2[0:y_longest-longest]+str2[y_longest:]

        return (new_str1, new_str2), longest

    def lcs_apply(x):
        """
        lcs_apply(x)

        A helper function that is applied to each pair of records
        in s1 and s2. Assigns a similarity score to each pair,
        between 0 and 1. Used by the pandas.apply method.

        Parameters
        ----------
        x : pandas.Series
          A pandas Series containing two strings to be compared.

        Returns
        -------
        Float
          A normalized similarity score.
        """
        if pandas.isnull(x[0]) or pandas.isnull(x[1]):
            return np.nan

        # Compute lcs value with first ordering.
        lcs_acc_1 = 0
        new_x_1 = (x[0], x[1])
        while True:
            # Get new string pair (iter_x) and length (iter_lcs)
            # for this iteration.
            iter_x, iter_lcs = lcs_iteration(new_x_1)
            if iter_lcs < min_len:
                # End if the longest substring is below the threshold
                break
            else:
                # Otherwise, accumulate length and start a new iteration
                # with the new string pair.
                new_x_1 = iter_x
                lcs_acc_1 = lcs_acc_1 + iter_lcs

        # Compute lcs value with second ordering.
        lcs_acc_2 = 0
        new_x_2 = (x[1], x[0])
        while True:
            # Get new string pair (iter_x) and length (iter_lcs)
            # for this iteration.
            iter_x, iter_lcs = lcs_iteration(new_x_2)
            if iter_lcs < min_len:
                # End if the longest substring is below the threshold
                break
            else:
                # Otherwise, accumulate length and start a new iteration
                # with the new string pair.
                new_x_2 = iter_x
                lcs_acc_2 = lcs_acc_2 + iter_lcs

        def normalize_lcs(lcs_value):
            """
            normalize_lcs(lcs_value)

            A helper function used to normalize the score produced by
            compute_score() to a score between 0 and 1. Applies one of
            the normalization schemes described in in Christen, Peter (2012).
            The normalization method is determined by the norm argument provided
            to the parent, longest_common_substring_similarity function.

            Parameters
            ----------
            lcs_value : Float
                The raw lcs length.

            Returns
            -------
            Float
                The normalized lcs length.
            """
            if len(x[0]) == 0 or len(x[1]) == 0:
                return 0
            if norm == 'overlap':
                return lcs_value / min(len(x[0]), len(x[1]))
            elif norm == 'jaccard':
                return lcs_value / (len(x[0])+len(x[1])-abs(lcs_value))
            elif norm == 'dice':
                return lcs_value*2 / (len(x[0])+len(x[1]))
            else:
                warnings.warn('Unrecognized longest common substring normalization. Defaulting to "dice" method.')
                return lcs_value*2 / (len(x[0])+len(x[1]))

        # Average the two orderings, since lcs may be sensitive to comparison order.
        return (normalize_lcs(lcs_acc_1)+normalize_lcs(lcs_acc_2)) / 2

    return conc.apply(lcs_apply)



from __future__ import division

import warnings
from functools import wraps

import pandas
import numpy

from recordlinkage.base import BaseIndexator
from recordlinkage.utils import IndexError
from recordlinkage.utils import VisibleDeprecationWarning
from recordlinkage.utils import merge_dicts
from recordlinkage.algorithms.string import qgram_similarity
from recordlinkage.utils import listify
from recordlinkage.measures import reduction_ratio
from recordlinkage.measures import full_index_size
from recordlinkage.algorithms.indexing import \
    random_pairs_with_replacement
from recordlinkage.algorithms.indexing import \
    random_pairs_without_replacement_small_frames
from recordlinkage.algorithms.indexing import \
    random_pairs_without_replacement_large_frames

from recordlinkage import rl_logging as logging


def check_index_names(func):
    # decorator to prevent index name conflicts. Used in functions like
    # blocking and SNI. Also useful for user defined functions.

    @wraps(func)
    def index_name_checker(df_a, df_b, *args, **kwargs):

        if (df_a.index.name is None or
                df_b.index.name is None or
                df_a.index.name is df_b.index.name or
                df_a.index.name in df_a.columns.tolist() or
                df_b.index.name in df_b.columns.tolist()):

            df_a_index_name = df_a.index.name
            df_b_index_name = df_b.index.name

            # temp update of the index name in each datatframe
            df_a.index.name = _save_label(list(df_a), s='index_x')
            df_b.index.name = _save_label(list(df_b), s='index_y')

            pairs = func(df_a, df_b, *args, **kwargs)

            # update the name of the index
            pairs.names = [df_a_index_name, df_b_index_name]

            df_a.index.name = df_a_index_name
            df_b.index.name = df_b_index_name

            return pairs

        else:

            return func(df_a, df_b, *args, **kwargs)

    return index_name_checker


def _save_label(labels, s='label'):

    i = 0

    while s in list(labels):

        i = i + 1
        s = s + '_' + str(i)

    return s

###########################
#       Algorithms        #
###########################


def _random_large_dedup(df_a, n, random_state=None):

    numpy.random.seed(random_state)

    n_max = full_index_size(df_a)

    if not isinstance(n, int) or n <= 0 or n > n_max:
        raise ValueError("n must be a integer satisfying 0<n<=%s" % n_max)

    full_index = _fullindex_dedup(df_a)
    sample = numpy.random.choice(
        numpy.arange(len(full_index)), n, replace=False
    )

    return full_index[sample]


def _random_large_link(df_a, df_b, n):

    n_max = full_index_size(df_a, df_b)

    if not isinstance(n, int) or n <= 0 or n > n_max:
        raise ValueError("n must be a integer satisfying 0<n<=%s" % n_max)

    full_index = _fullindex_link(df_a, df_b)
    sample = numpy.random.choice(
        numpy.arange(len(full_index)), n, replace=False
    )

    return full_index[sample]


def _random_small_link(df_a, df_b, n):

    n_max = full_index_size(df_a, df_b)

    if not isinstance(n, int) or n <= 0 or n > n_max:
        raise ValueError("n must be a integer satisfying 0<n<=%s" % n_max)

    levels = [df_a.index.values, df_b.index.values]
    names = [df_a.index.name, df_b.index.name]

    # Initialize pandas MultiIndex
    pairs = pandas.MultiIndex(levels=levels, labels=[[], []], names=names)

    # Run as long as the number of pairs is less than the requested number
    # of pairs n.
    while len(pairs) < n:

        # The number of pairs to sample (sample twice as much record pairs
        # because the duplicates are dropped).
        n_sample = (n - len(pairs)) * 2
        sample_a = numpy.random.randint(len(df_a), size=n_sample)
        sample_b = numpy.random.randint(len(df_b), size=n_sample)

        # Make a pandas MultiIndex of the sample above
        pairs_sample = pandas.MultiIndex(
            levels=levels, labels=[sample_a, sample_b], names=names
        )

        pairs = pairs.append(pairs_sample).drop_duplicates()

    return pairs[0:n]


def _eye(df_a, df_b):

    return pandas.MultiIndex.from_arrays(
        [df_a.index.values, df_b[0:len(df_a)].index.values],
        names=[df_a.index.name, df_b.index.name]
    )


def _fullindex_link(df_a, df_b):

    return pandas.MultiIndex.from_product(
        [df_a.index.values, df_b.index.values],
        names=[df_a.index.name, df_b.index.name]
    )


def _fullindex_dedup(df_a):

    return pandas.MultiIndex(
        levels=[df_a.index.values, df_a.index.values],
        labels=numpy.triu_indices(len(df_a.index), k=1),
        names=[df_a.index.name, df_a.index.name],
        verify_integrity=False
    )


@check_index_names
def _blockindex(df_a, df_b, on=None, left_on=None, right_on=None):

    # Index name conflicts do not occur. They are handled in the decorator.

    if on:
        left_on, right_on = on, on

    # Rows with missing values on the on attributes are dropped.
    data_left = df_a.dropna(axis=0, how='any', subset=numpy.unique(left_on))
    data_right = df_b.dropna(axis=0, how='any', subset=numpy.unique(right_on))

    # Join
    pairs = data_left.reset_index().merge(
        data_right.reset_index(),
        how='inner',
        left_on=left_on,
        right_on=right_on,
    ).set_index([df_a.index.name, df_b.index.name])

    return pairs.index


@check_index_names
def _sortedneighbourhood(
        df_a, df_b, column, window=3, sorting_key_values=None,
        block_on=[], block_left_on=[], block_right_on=[]):

    # Check if window is an odd number
    if not isinstance(window, int) or (window < 0) or not bool(window % 2):
        raise ValueError(
            'The given window length is not a positive and odd integer.')

    block_on = [block_on] if type(block_on) != list else block_on
    block_left_on = [block_left_on] if type(
        block_left_on) != list else block_left_on
    block_right_on = [block_right_on] if type(
        block_right_on) != list else block_right_on

    block_left_on, block_right_on = [
        block_on, block_on] if block_on else ([], [])
    keys_left = [column] + block_left_on
    keys_right = [column] + block_right_on

    df_a = df_a[df_a[column].notnull()]  # df_a.dropna(inplace=True)
    df_b = df_b[df_b[column].notnull()]  # df_a.dropna(inplace=True)

    # sorting_key_values is the terminology in Data Matching [Christen, 2012]
    if sorting_key_values is None:

        # Combine the results
        sorting_key_values = numpy.sort(numpy.unique(
            numpy.concatenate([df_a[column].values, df_b[column].values])
        ))

    sorting_key_factors = numpy.arange(len(sorting_key_values))

    data_dict_A = {kl: df_a[kl] for kl in keys_left}
    data_dict_B = {kl: df_b[kl] for kl in keys_right}

    sorted_index = pandas.Series(
        index=sorting_key_values, data=sorting_key_factors)
    sorted_df_A = pandas.DataFrame(
        merge_dicts(
            data_dict_A,
            {column: df_a[column].map(sorted_index),
             df_a.index.name: df_a.index.values}))
    sorted_df_B = pandas.DataFrame(
        {column: df_b[column].map(sorted_index),
            df_b.index.name: df_b.index.values})

    pairs_concat = None

    # Internal window size
    _window = int((window - 1) / 2)

    for w in range(-_window, _window + 1):

        df = pandas.DataFrame(
            merge_dicts(
                data_dict_B,
                {
                    column: sorted_df_B[column] + w,
                    df_b.index.name: df_b.index.values
                }
            )
        )

        pairs = sorted_df_A.merge(
            df, left_on=keys_left, right_on=keys_right, how='inner'
        ).set_index(
            [df_a.index.name, df_b.index.name]
        )

        if pairs_concat is None:
            pairs_concat = pairs.index
        else:
            pairs_concat = pairs.index.append(pairs_concat)

    return pairs_concat


def _qgram(df_a, df_b, on=None, left_on=None, right_on=None, threshold=0.8):

    if on:
        left_on, right_on = on, on

        # Rows with missing values on the on attributes are dropped.

    fi = pandas.MultiIndex.from_product(
        [df_a[left_on].dropna(axis=0).index.values,
         df_b[right_on].dropna(axis=0).index.values],
        names=[df_a.index.name, df_b.index.name]
    )

    if len(fi) > 0:
        bool_index = qgram_similarity(
            df_a.loc[fi.get_level_values(0), left_on],
            df_b.loc[fi.get_level_values(1), right_on]
        ) >= threshold

        return fi[bool_index]

    else:
        return fi


class PairsCore(object):
    """Core class for making record pairs.

    """

    def __init__(self, df_a, df_b=None, chunks=None, verify_integrity=True):

        warnings.warn(
            "indexing api changed, see the documentation for the new format",
            VisibleDeprecationWarning
        )

        self.df_a = df_a
        self.df_b = df_b
        self.chunks = chunks

        self.deduplication = True if df_b is None else False

        if verify_integrity:

            if not self.df_a.index.is_unique:
                raise IndexError('index of DataFrame df_a is not unique.')

            if not self.deduplication and not self.df_b.index.is_unique:
                raise IndexError('index of DataFrame df_b is not unique.')

        self.n_pairs = 0

    @property
    def maximum_number_of_pairs(self):
        """ the maximum number of record pairs """

        if self.deduplication:
            return full_index_size(self.df_a)
        else:
            return full_index_size(self.df_a, self.df_b)

    # -- Index methods ------------------------------------------------------

    def index(self, index_func, *args, **kwargs):
        """Main function to make an index of record pairs.

        Use a custom function to make record pairs of one or two dataframes.
        Each function should return a pandas.MultiIndex with record pairs.

        Parameters
        ----------
        index_func: function
            An indexing function

        Returns
        -------
        pandas.MultiIndex
            A pandas multiindex with record pairs. The first position in the
            index is the index of a record in dataframe A and the second
            position is the index of a record in dataframe B.

        """

        # linking
        if not self.deduplication:

            pairs = index_func(self.df_a, self.df_b, *args, **kwargs)

        # deduplication
        else:

            if index_func.__name__.endswith('_dedup'):

                pairs = index_func(self.df_a, *args, **kwargs)

            else:

                pairs = index_func(self.df_a, self.df_a.copy(),
                                   *args, **kwargs)
                # Remove all double pairs!
                pairs = pairs[
                    pairs.get_level_values(0) < pairs.get_level_values(1)
                ]

        # If there are no chunks, then use the first item of the generator
        if self.chunks is None:
            return pairs

        # Use the chunks
        else:
            return self._iter_pairs(pairs)

    def _iter_pairs(self, pairs):

        if not isinstance(self.chunks, int):
            raise ValueError('argument chunks needs to be integer type')

        bins = numpy.arange(0, len(pairs), step=self.chunks)

        for b in bins:

            yield pairs[b:b + self.chunks]


class Pairs(PairsCore):
    """Make record pairs.

    This class can be used to make record pairs. Multiple indexation methods
    can be used to make a smart selection of record pairs. Indexation methods
    included:

    - Full indexing
    - Blocking
    - Sorted Neighbourhood
    - Random indexing
    - Q-gram indexing

    For more information about indexing and methods to reduce the number of
    record pairs, see Christen 2012 [christen2012]_.

    Parameters
    ----------
    df_a : pandas.DataFrame
        The first dataframe.
    df_b : pandas.DataFrame
        The second dataframe.
    chunks : int
        The chunks to divide the result in.
    verify_integrity : bool
        Verify the integrity of the dataframes
            (default True).

    Attributes
    ----------
    df_a : pandas.DataFrame
        The first DataFrame.
    df_b : pandas.DataFrame
        The second DataFrame.
    n_pairs : int
        The number of candidate record pairs.
    reduction : float
        The reduction ratio.

    Examples
    --------
    In the following example, the record pairs are made for two historical
    datasets with census data. The datasets are named ``census_data_1980``
    and ``census_data_1990``.

    >>> pcl = recordlinkage.Pairs(census_data_1980, census_data_1990)
    >>> pcl.block('first_name')

    References
    ----------
    .. [christen2012] Christen, 2012. Data Matching Concepts and
            Techniques for Record Linkage, Entity Resolution, and
            Duplicate Detection

    """

    def full(self):
        """Make a full index (all possible pairs).

        Make an index with all possible record pairs. In case of linking two
        dataframes (A and B), the number of pairs is len(A)*len(B). In case of
        deduplicating a dataframe A, the number of pairs is
        len(A)*(len(A)-1)/2.

        Returns
        -------
        pandas.MultiIndex
            The index of the candidate record pairs

        """
        if self.deduplication:
            return self.index(_fullindex_dedup)
        else:
            return self.index(_fullindex_link)

    def block(self, on=None, left_on=None, right_on=None):
        """Make a block index.

        Return all record pairs that agree on the passed attribute(s). This
        method is known as *blocking*

        Parameters
        ----------
        on : label
            A column name or a list of column names. These columns are used to
            block on.
        left_on : label
            A column name or a list of column names of dataframe A. These
            columns are used to block on.
        right_on : label
            A column name or a list of column names of dataframe B. These
            columns are used to block on.

        Returns
        -------
        pandas.MultiIndex
            The index of the candidate record pairs

        """
        return self.index(
            _blockindex,
            on=on, left_on=left_on, right_on=right_on
        )

    def sortedneighbourhood(
        self, on, window=3, sorting_key_values=None, block_on=[],
        block_left_on=[], block_right_on=[]
    ):
        """Make a Sorted Neighbourhood index.

        Parameters
        ----------
        on: label
            Specify the on to make a sorted index
        window: int
            The width of the window, default is 3
        sorting_key_values: array
            A list of sorting key values (optional).
        block_on: label
            Additional columns to use standard blocking on
        block_left_on: label
            Additional columns of the left dataframe to use standard blocking
            on.
        block_right_on: label
            Additional columns of the right dataframe to use standard
            blocking on

        Returns
        -------
        pandas.MultiIndex
            The index of the candidate record pairs

        """

        return self.index(
            _sortedneighbourhood,
            on, window=window, sorting_key_values=sorting_key_values,
            block_on=block_on, block_left_on=block_left_on,
            block_right_on=block_right_on)

    def random(self, n):
        """Make an index of randomly selected record pairs.

        Parameters
        ----------
        n : int
            The number of record pairs to return. The integer n should satisfy
            0 < n <= len(A)*len(B).

        Returns
        -------
        pandas.MultiIndex
            The index of the candidate record pairs

        """

        if not isinstance(n, int):
            raise ValueError("an integer is required")

        if self.deduplication:
            return self.index(_random_large_dedup, n)
        else:
            if n < 0.1 * self.maximum_number_of_pairs:
                return self.index(_random_large_link, n)
            else:
                return self.index(_random_small_link, n)

    def qgram(self, *args, **kwargs):
        """
        qgram(on=None, left_on=None, right_on=None, threshold=0.8)

        Use Q-gram string comparing metric to make an index.

        Parameters
        ----------
        on : label
            A column name or a list of column names. These columns are used to
            index on.
        left_on : label
            A column name or a list of column names of dataframe A. These
            columns are used to index on.
        right_on : label
            A column name or a list of column names of dataframe B. These
            columns are used to index on.
        threshold : float
            Record pairs with a similarity above the threshold are candidate
            record pairs. [Default 0.8]

        Returns
        -------
        pandas.MultiIndex
            The index of the candidate record pairs

        """
        return self.index(_qgram, *args, **kwargs)

    def eye(self):
        # Only for internal use

        return self.index(_eye)

    @property
    def reduction(self):

        return 1 - self.n_pairs / self.maximum_number_of_pairs


##############################################################################
#
# This section contains the indexing classes for the new indexing API.
#


class FullIndex(BaseIndexator):
    """FullIndex()
    Class to generate a 'full' index.

    A full index is an index with all possible combinations of record pairs.
    In case of linking, this indexation method generates the cartesian product
    of both DataFrame's. In case of deduplicating DataFrame A, this indexation
    method are the pairs defined by the upper triangular matrix of the A x A.

    Note
    ----
    This indexation method can be slow for large DataFrame's. The number of
    comparisons scales quadratic.
    Also, not all classifiers work well with large numbers of record pairs
    were most of the pairs are distinct.

    """

    def __init__(self, *args, **kwargs):
        super(FullIndex, self).__init__(*args, **kwargs)

    def _link_index(self, df_a, df_b):

        n_max = full_index_size((df_a, df_b))

        if n_max > 1e7:
            logging.warn(
                "The number of record pairs is large. Consider a different "
                "indexation algorithm for better performance. "
            )

        return pandas.MultiIndex.from_product(
            [df_a.index.values, df_b.index.values],
            names=[df_a.index.name, df_b.index.name]
        )

    def _dedup_index(self, df_a):

        n_max = full_index_size((df_a))

        if n_max > 1e7:
            logging.warn(
                "The number of record pairs is large. Consider a different "
                "indexation algorithm for better performance. "
            )

        levels = [df_a.index.values, df_a.index.values]
        labels = numpy.triu_indices(len(df_a.index), k=1)
        names = [df_a.index.name, df_a.index.name]

        return pandas.MultiIndex(
            levels=levels,
            labels=labels,
            names=names,
            verify_integrity=False
        )


class BlockIndex(BaseIndexator):
    """BlockIndex(on=None, left_on=None, right_on=None)
    Make candidate record pairs that agree on one or more variables.

    Returns all record pairs that agree on the given variable(s). This method
    is known as *blocking*. Blocking is an effective way to make a subset of
    the record space (A * B).

    Parameters
    ----------
    on : label, optional
        A column name or a list of column names. These column(s) are used to
        block on. When linking two dataframes, the 'on' argument needs to be
        present in both dataframes.
    left_on : label, optional
        A column name or a list of column names of dataframe A. These
        columns are used to block on. This argument is ignored when argument
        'on' is given.
    right_on : label, optional
        A column name or a list of column names of dataframe B. These
        columns are used to block on. This argument is ignored when argument
        'on' is given.

    Examples
    --------
    In the following example, the record pairs are made for two historical
    datasets with census data. The datasets are named ``census_data_1980``
    and ``census_data_1990``.

    >>> indexer = recordlinkage.BlockIndex(on='first_name')
    >>> indexer.index(census_data_1980, census_data_1990)

    """

    def __init__(self, on=None, left_on=None, right_on=None,
                 *args, **kwargs):
        super(BlockIndex, self).__init__(*args, **kwargs)

        # variables to block on
        self.on = on
        self.left_on = left_on
        self.right_on = right_on

    def __repr__(self):

        class_name = self.__class__.__name__

        if self.on is not None:
            left_on = right_on = self.on
        else:
            left_on, right_on = self.left_on, self.right_on

        return "<{} left_on={!r}, right_on={!r}>".format(
            class_name, left_on, right_on)

    def _link_index(self, df_a, df_b):

        if self.on is not None:
            if self.left_on is not None or self.right_on is not None:
                raise IndexError('Can only pass argument "on" OR "left_on" '
                                 'and "right_on", not a combination of both.')
            left_on = right_on = listify(self.on)
        else:
            if self.left_on is None and self.right_on is None:
                raise IndexError('pass argument "on" OR "left_on" and '
                                 '"right_on" at class initalization.')
            elif self.left_on is None:
                raise IndexError('Argument "left_on" is missing '
                                 'at class initalization.')
            elif self.right_on is None:
                raise IndexError('Argument "right_on" is missing '
                                 'at class initalization.')
            else:
                left_on = listify(self.left_on)
                right_on = listify(self.right_on)

        blocking_keys = ["blocking_key_%d" % i for i, v in enumerate(left_on)]

        # make a dataset for the data on the left
        data_left = df_a[left_on].dropna(axis=0, how='any', inplace=False)
        data_left.columns = blocking_keys
        data_left['index_x'] = data_left.index

        # make a dataset for the data on the right
        data_right = df_b[right_on].dropna(axis=0, how='any', inplace=False)
        data_right.columns = blocking_keys
        data_right['index_y'] = data_right.index

        # merge the dataframes
        pairs = data_left.merge(
            data_right, how='inner', on=blocking_keys
        ).set_index(['index_x', 'index_y'])

        return pairs.index.rename([df_a.index.name, df_b.index.name])


class SortedNeighbourhoodIndex(BaseIndexator):
    """SortedNeighbourhoodIndex(on=None, left_on=None, right_on=None, window=3, sorting_key_values=None, block_on=[], block_left_on=[], block_right_on=[])
    Make candidate record pairs with the SortedNeighbourhood algorithm.

    This algorithm returns record pairs that agree on the sorting key, but
    also records pairs in their neighbourhood. A large window size results in
    more record pairs. A window size of 1 returns the blocking index.

    The Sorted Neighbourhood Index method is a great method when there is
    relatively large amount of spelling mistakes. Blocking will fail in that
    situation because it excludes to many records on minor spelling mistakes.

    Parameters
    ----------
    on : label, optional
        The column name of the sorting key. When linking two dataframes, the
        'on' argument needs to be present in both dataframes.
    left_on : label, optional
        The column name of the sorting key of the first/left dataframe. This
        argument is ignored when argument 'on' is not None.
    right_on : label, optional
        The column name of the sorting key of the second/right dataframe. This
        argument is ignored when argument 'on' is not None.
    window: int, optional
        The width of the window, default is 3
    sorting_key_values: array, optional
        A list of sorting key values (optional).
    block_on: label
        Additional columns to apply standard blocking on.
    block_left_on: label
        Additional columns in the left dataframe to apply standard blocking
        on.
    block_right_on: label
        Additional columns in the right dataframe to apply standard blocking
        on.


    Examples
    --------
    In the following example, the record pairs are made for two historical
    datasets with census data. The datasets are named ``census_data_1980``
    and ``census_data_1990``.

    >>> indexer = recordlinkage.SortedNeighbourhoodIndex('first_name', window=9)
    >>> indexer.index(census_data_1980, census_data_1990)

    When the sorting key has different names in both dataframes:

    >>> indexer = recordlinkage.SortedNeighbourhoodIndex(
            left_on='first_name', right_on='given_name', window=9
        )
    >>> indexer.index(census_data_1980, census_data_1990)

    """

    def __init__(self, on=None, left_on=None, right_on=None, window=3,
                 sorting_key_values=None, block_on=[], block_left_on=[],
                 block_right_on=[], *args, **kwargs):
        super(SortedNeighbourhoodIndex, self).__init__(*args, **kwargs)

        # variables to block on
        self.on = on
        self.left_on = left_on
        self.right_on = right_on
        self.window = window
        self.sorting_key_values = sorting_key_values
        self.block_on = block_on
        self.block_left_on = block_left_on
        self.block_right_on = block_right_on

    def __repr__(self):

        class_name = self.__class__.__name__

        if self.on is not None:
            left_on = right_on = self.on
        else:
            left_on, right_on = self.left_on, self.right_on

        return "<{} left_on={!r}, right_on={!r}>".format(
            class_name, left_on, right_on)

    def _get_sorting_key_values(self, array1, array2):
        """return the sorting key values as a series"""

        concat_arrays = numpy.concatenate([array1, array2])
        unique_values = numpy.unique(concat_arrays)

        return numpy.sort(unique_values)

    def _link_index(self, df_a, df_b):

        if self.on is not None:
            if self.left_on is not None or self.right_on is not None:
                raise IndexError('Can only pass argument "on" OR "left_on" '
                                 'and "right_on", not a combination of both.')
            left_on = right_on = listify(self.on)
        else:
            if self.left_on is None and self.right_on is None:
                raise IndexError('pass argument "on" OR "left_on" and '
                                 '"right_on" at class initalization.')
            elif self.left_on is None:
                raise IndexError('Argument "left_on" is missing '
                                 'at class initalization.')
            elif self.right_on is None:
                raise IndexError('Argument "right_on" is missing '
                                 'at class initalization.')
            else:
                left_on = listify(self.left_on)
                right_on = listify(self.right_on)

        window = self.window

        # Check if window is an odd number
        if not isinstance(window, int) or (window < 0) or not bool(window % 2):
            raise ValueError(
                'window is not a positive and odd integer')

        # # sorting key is single column
        # if isinstance(self.on, (tuple, list, dict)):
        #     raise ValueError(
        #         "sorting key is not a label")

        # make blocking keys correct

        block_left_on = listify(self.block_left_on)
        block_right_on = listify(self.block_right_on)

        if self.block_on:
            block_left_on = listify(self.block_on)
            block_right_on = listify(self.block_on)

        # drop missing values and columns without relevant information
        data_left = df_a[listify(left_on) + block_left_on].dropna(
            axis=0, how='any', inplace=False
        )
        data_left.columns = ['sorting_key'] + \
            ["blocking_key_%d" % i for i, v in enumerate(block_left_on)]
        data_left['index_x'] = data_left.index

        data_right = df_b[listify(right_on) + block_right_on].dropna(
            axis=0, how='any', inplace=False
        )
        data_right.columns = ['sorting_key'] + \
            ["blocking_key_%d" % i for i, v in enumerate(block_right_on)]
        data_right['index_y'] = data_right.index

        # sorting_key_values is the terminology in Data Matching [Christen,
        # 2012]
        if self.sorting_key_values is None:

            self.sorting_key_values = self._get_sorting_key_values(
                data_left['sorting_key'].values,
                data_right['sorting_key'].values
            )

        sorting_key_factors = pandas.Series(
            numpy.arange(len(self.sorting_key_values)),
            index=self.sorting_key_values)

        data_left['sorting_key'] = data_left[
            'sorting_key'].map(sorting_key_factors)
        data_right['sorting_key'] = data_right[
            'sorting_key'].map(sorting_key_factors)

        # Internal window size
        _window = int((window - 1) / 2)

        def merge_lagged(x, y, w):
            """Merge two dataframes with a lag on in the sorting key."""

            y = y.copy()
            y['sorting_key'] = y['sorting_key'] + w

            return x.merge(y, how='inner')

        pairs_concat = [merge_lagged(data_left, data_right, w)
                        for w in range(-_window, _window + 1)]

        pairs = pandas.concat(pairs_concat, axis=0).set_index(
            ['index_x', 'index_y']
        ).index.rename([df_a.index.name, df_b.index.name])

        return pairs


class RandomIndex(BaseIndexator):
    """RandomIndex(n, replace=True, random_state=None)
    Class to generate random pairs of records.

    This class returns random pairs of records with or without replacement.
    Use the random_state parameter to seed the algorithm and reproduce
    results. This way to make record pairs is useful for the training of
    unsupervised learning models for record linkage.

    Parameters
    ----------
    n : int
        The number of record pairs to return. In case replace=False, the
        integer n should be bounded by 0 < n <= n_max where n_max is the
        maximum number of pairs possible.
    replace : bool, optional
        Whether the sample of record pairs is with or without replacement.
        Default: True
    random_state : int or numpy.random.RandomState, optional
        Seed for the random number generator (if int), or numpy RandomState
        object.

    """

    def __init__(self, n, replace=True, random_state=None, *args, **kwargs):
        super(RandomIndex, self).__init__(*args, **kwargs)

        self.n = n
        self.replace = replace
        self.random_state = random_state

    def __repr__(self):

        class_name = self.__class__.__name__

        return "<{} n={!r}, replace={!r}>".format(
            class_name, self.n, self.replace)

    def _link_index(self, df_a, df_b):

        shape = (len(df_a), len(df_b))
        n_max = full_index_size(shape)

        if not isinstance(self.n, int):
            raise ValueError('n must be an integer')

        # with replacement
        if self.replace:

            if n_max == 0:
                raise ValueError(
                    "one of the dataframes is empty")

            pairs = random_pairs_with_replacement(
                self.n, shape, self.random_state)

        # without replacement
        else:

            if self.n <= 0 or self.n > n_max:
                raise ValueError(
                    "n must be a integer satisfying 0<n<=%s" % n_max)

            # large dataframes
            if n_max < 1e6:
                pairs = random_pairs_without_replacement_small_frames(
                    self.n, shape, self.random_state)
            # small dataframes
            else:
                pairs = random_pairs_without_replacement_large_frames(
                    self.n, shape, self.random_state)

        levels = [df_a.index.values, df_b.index.values]
        labels = pairs
        names = [df_a.index.name, df_b.index.name]

        return pandas.MultiIndex(
            levels=levels,
            labels=labels,
            names=names,
            verify_integrity=False
        )

    def _dedup_index(self, df_a):

        shape = (len(df_a),)

        # with replacement
        if self.replace:
            pairs = random_pairs_with_replacement(
                self.n, shape, self.random_state)

        # without replacement
        else:

            n_max = full_index_size(shape)

            if not isinstance(self.n, int) or self.n <= 0 or self.n > n_max:
                raise ValueError(
                    "n must be a integer satisfying 0<n<=%s" % n_max)

            # large dataframes
            if n_max < 1e6:
                pairs = random_pairs_without_replacement_small_frames(
                    self.n, shape, self.random_state)
            # small dataframes
            else:
                pairs = random_pairs_without_replacement_large_frames(
                    self.n, shape, self.random_state)

        levels = [df_a.index.values, df_a.index.values]
        labels = pairs
        names = [df_a.index.name, df_a.index.name]

        return pandas.MultiIndex(
            levels=levels,
            labels=labels,
            names=names,
            verify_integrity=False
        )



from __future__ import division
from __future__ import unicode_literals

from functools import wraps

import pandas
import numpy as np

from recordlinkage.base import BaseCompare
from recordlinkage.types import is_list_like
from recordlinkage.algorithms.compare import _compare_exact
from recordlinkage.algorithms.compare import _compare_dates
from recordlinkage.algorithms.distance import _1d_distance
from recordlinkage.algorithms.distance import _haversine_distance
from recordlinkage.algorithms.numeric import _step_sim
from recordlinkage.algorithms.numeric import _linear_sim
from recordlinkage.algorithms.numeric import _squared_sim
from recordlinkage.algorithms.numeric import _exp_sim
from recordlinkage.algorithms.numeric import _gauss_sim
from recordlinkage.algorithms.string import jaro_similarity
from recordlinkage.algorithms.string import jarowinkler_similarity
from recordlinkage.algorithms.string import levenshtein_similarity
from recordlinkage.algorithms.string import damerau_levenshtein_similarity
from recordlinkage.algorithms.string import qgram_similarity
from recordlinkage.algorithms.string import cosine_similarity
from recordlinkage.algorithms.string import smith_waterman_similarity
from recordlinkage.algorithms.string import longest_common_substring_similarity


def fillna_decorator(missing_value=np.nan):

    def real_decorator(func):

        @wraps(func)
        def func_wrapper(*args, **kwargs):

            mv = kwargs.pop('missing_value', missing_value)

            result = func(*args, **kwargs)

            # fill missing values if missing_value is not a missing value like
            # NaN or None.
            if pandas.notnull(mv):
                if isinstance(result, (np.ndarray)):
                    result[np.isnan(result)] = mv
                else:
                    result.fillna(mv, inplace=True)

            return result

        return func_wrapper

    return real_decorator


def _check_labels(labels, df):

    labels = [labels] if not is_list_like(labels) else labels
    cols = df.columns.tolist() if isinstance(df, pandas.DataFrame) else df

    # Do some checks
    for label in labels:
        if label not in cols:
            raise KeyError(
                'label [%s] not in dataframe' % label
            )


@fillna_decorator(0)
def _string_internal(s1, s2, call_method, threshold=None, *args, **kw):

    c = call_method(s1, s2, *args, **kw)

    if threshold:
        return (c >= threshold).astype(np.float64)
    else:
        return c


@fillna_decorator(0)
def _num_internal(s1, s2, call_method, *args, **kwargs):
    """Internal function to compute the numeric similarity."""

    # compute the 1D distance between the values
    d = _1d_distance(s1, s2)

    return call_method(d, *args, **kwargs)


@fillna_decorator(0)
def _geo_internal(lat1, lng1, lat2, lng2, call_method, *args, **kw):

    # compute the 1D distance between the values
    d = _haversine_distance(lat1, lng1, lat2, lng2)

    return call_method(d, *args, **kw)


@fillna_decorator(0)
def _dates_internal(s1, s2, *args, **kwargs):

    return _compare_dates(s1, s2, *args, **kwargs)


class Compare(BaseCompare):
    """Compare(n_jobs=1, indexing_type='label')

    Class to compare record pairs with efficiently.

    Class to compare the attributes of candidate record pairs. The ``Compare``
    class has methods like ``string``, ``exact`` and ``numeric`` to initialise
    the comparing of the records. The ``compute`` method is used to start the
    actual comparing.

    Parameters
    ----------
    n_jobs : integer, optional (default=1)
        The number of jobs to run in parallel for comparing of record pairs.
        If -1, then the number of jobs is set to the number of cores.
    indexing_type : string, optional (default='label')
        The indexing type. The MultiIndex is used to index the DataFrame(s).
        This can be done with pandas ``.loc`` or with ``.iloc``. Use the value
        'label' to make use of ``.loc`` and 'position' to make use of
        ``.iloc``. The value 'position' is only available when the MultiIndex
        consists of integers. The value 'position' is much faster.

    Example
    -------
    Consider two historical datasets with census data to link. The datasets
    are named ``census_data_1980`` and ``census_data_1990``. The MultiIndex
    ``candidate_pairs`` contains the record pairs to compare. The record pairs
    are compared on the first name, last name, sex, date of birth, address,
    place, and income::

        # initialise class
        comp = recordlinkage.Compare()

        # initialise similarity measurement algorithms
        comp.string('first_name', 'name', method='jarowinkler')
        comp.string('lastname', 'lastname', method='jarowinkler')
        comp.exact('dateofbirth', 'dob')
        comp.exact('sex', 'sex')
        comp.string('address', 'address', method='levenshtein')
        comp.exact('place', 'place')
        comp.numeric('income', 'income')

        # the method .compute() returns the DataFrame with the feature vectors.
        comp.compute(candidate_pairs, census_data_1980, census_data_1990)

    """

    def exact(self, s1, s2, *args, **kwargs):
        """
        exact(s1, s2, agree_value=1, disagree_value=0, missing_value=0, label=None)

        Compare the record pairs exactly.

        This method initialises the exact similarity measurement between
        values. The similarity is 1 in case of agreement and 0 otherwise.

        Parameters
        ----------

        s1 : str or int
            Field name to compare in left DataFrame.
        s2 : str or int
            Field name to compare in right DataFrame.
        agree_value : float, str, numpy.dtype
            The value when two records are identical. Default 1. If 'values'
            is passed, then the value of the record pair is passed.
        disagree_value : float, str, numpy.dtype
            The value when two records are not identical.
        missing_value : float, str, numpy.dtype
            The value for a comparison with a missing value. Default 0.
        label : label
            The label of the column in the resulting dataframe.

        """

        label = kwargs.pop('label', None)

        return self._compare_vectorized(
            _compare_exact, s1, s2, args, kwargs,
            label=label,
            name="exact",
            description="Compare record pairs exactly."
        )

    def string(self, s1, s2, method='levenshtein', threshold=None,
               *args, **kwargs):
        """
        string(s1, s2, method='levenshtein', threshold=None, missing_value=0, label=None)

        Compute the (partial) similarity between strings values.

        This method initialises the similarity measurement between string
        values. The implemented algorithms are: 'jaro','jarowinkler',
        'levenshtein', 'damerau_levenshtein', 'qgram' or 'cosine'. In case of
        agreement, the similarity is 1 and in case of complete disagreement it
        is 0. The Python Record Linkage Toolkit uses the `jellyfish` package
        for the Jaro, Jaro-Winkler, Levenshtein and Damerau-Levenshtein
        algorithms.

        Parameters
        ----------
        s1 : str or int
            The name or position of the column in the left DataFrame.
        s2 : str or int
            The name or position of the column in the right DataFrame.
        method : str, default 'levenshtein'
            An approximate string comparison method. Options are ['jaro',
            'jarowinkler', 'levenshtein', 'damerau_levenshtein', 'qgram',
            'cosine', 'smith_waterman', 'lcs']. Default: 'levenshtein'
        threshold : float, tuple of floats
            A threshold value. All approximate string comparisons higher or
            equal than this threshold are 1. Otherwise 0.
        missing_value : numpy.dtype
            The value for a comparison with a missing value. Default 0.
        label : label
            The label of the column in the resulting dataframe.

        """

        if method == 'jaro':
            str_sim_alg = jaro_similarity

        elif method in ['jarowinkler', 'jaro_winkler', 'jw']:
            str_sim_alg = jarowinkler_similarity

        elif method == 'levenshtein':
            str_sim_alg = levenshtein_similarity

        elif method in ['dameraulevenshtein', 'damerau_levenshtein', 'dl']:
            str_sim_alg = damerau_levenshtein_similarity

        elif method == 'q_gram' or method == 'qgram':
            str_sim_alg = qgram_similarity

        elif method == 'cosine':
            str_sim_alg = cosine_similarity

        elif method in ['smith_waterman', "smithwaterman", "sw"]:
            str_sim_alg = smith_waterman_similarity

        elif method in ['longest_common_substring', 'lcs']:
            str_sim_alg = longest_common_substring_similarity

        else:
            raise ValueError(
                "The algorithm '{}' is not known.".format(method))

        label = kwargs.pop('label', None)
        args = (str_sim_alg, threshold) + args

        return self._compare_vectorized(
            _string_internal, s1, s2, args, kwargs,
            label=label,
            name="string '{}'".format(method),
            description="Compare record pairs on string with '{}'"
                        "algorithm".format(method)
        )

    def numeric(self, s1, s2, method='linear', *args, **kwargs):
        """
        numeric(s1, s2, method='linear', offset, scale, origin=0, missing_value=0, label=None)

        Compute the (partial) similarity between numeric values.

        This method initialises the similarity measurement between numeric
        values. The implemented algorithms are: 'step', 'linear', 'exp',
        'gauss' or 'squared'. In case of agreement, the similarity is 1 and in
        case of complete disagreement it is 0. The implementation is similar
        with numeric comparing in ElasticSearch, a full-text search tool. The
        parameters are explained in the image below (source ElasticSearch, The
        Definitive Guide)

        .. image:: /images/elas_1705.png
            :width: 100%
            :target: https://www.elastic.co/guide/en/elasticsearch/guide/current/decay-functions.html
            :alt: Decay functions, like in ElasticSearch

        Parameters
        ----------
        s1 : str or int
            The name or position of the column in the left DataFrame.
        s2 : str or int
            The name or position of the column in the right DataFrame.
        method : float
            The metric used. Options 'step', 'linear', 'exp', 'gauss' or
            'squared'. Default 'linear'.
        offset : float
            The offset. See image above.
        scale : float
            The scale of the numeric comparison method. See the image above.
            This argument is not available for the 'step' algorithm.
        origin : str
            The shift of bias between the values. See image above.
        missing_value : numpy.dtype
            The value if one or both records have a missing value on the
            compared field. Default 0.
        label : label
            The label of the column in the resulting dataframe.

        Note
        ----
        Numeric comparing can be an efficient way to compare date/time
        variables. This can be done by comparing the timestamps.

        """

        if method == 'step':
            num_sim_alg = _step_sim

        elif method in ['linear', 'lin']:
            num_sim_alg = _linear_sim

        elif method == 'squared':
            num_sim_alg = _squared_sim

        elif method in ['exp', 'exponential']:
            num_sim_alg = _exp_sim

        elif method in ['gauss', 'gaussian']:
            num_sim_alg = _gauss_sim

        else:
            raise ValueError(
                "The algorithm '{}' is not known.".format(method))

        label = kwargs.pop('label', None)
        args = (num_sim_alg,) + args

        return self._compare_vectorized(
            _num_internal, s1, s2, args, kwargs,
            label=label,
            name="numeric '{}'".format(method),
            description="Compare record pairs on numeric with '{}'"
                        "algorithm".format(method)
        )

    def geo(self, lat1, lng1, lat2, lng2, method='linear', *args, **kwargs):
        """
        geo(lat1, lng1, lat2, lng2, method='linear', offset, scale, origin=0, missing_value=0, label=None)

        Compute the (partial) similarity between WGS84 coordinate values.

        Compare the geometric (haversine) distance between two WGS-
        coordinates. The similarity algorithms are 'step', 'linear', 'exp',
        'gauss' or 'squared'. The similarity functions are the same as in
        :meth:`recordlinkage.comparing.Compare.numeric`

        Parameters
        ----------
        lat1 : str or int
            The name or position of the column in the left DataFrame.
        lng1 : str or int
            The name or position of the column in the left DataFrame.
        lat2 : str or int
            The name or position of the column in the right DataFrame.
        lng2 : str or int
            The name or position of the column in the right DataFrame.
        method : str
            The metric used. Options 'step', 'linear', 'exp', 'gauss' or
            'squared'. Default 'linear'.
        offset : float
            The offset. See Compare.numeric.
        scale : float
            The scale of the numeric comparison method. See Compare.numeric.
            This argument is not available for the 'step' algorithm.
        origin : float
            The shift of bias between the values. See Compare.numeric.
        missing_value : numpy.dtype
            The value for a comparison with a missing value. Default 0.
        label : label
            The label of the column in the resulting dataframe.

        """

        if method == 'step':
            num_sim_alg = _step_sim

        elif method in ['linear', 'lin']:
            num_sim_alg = _linear_sim

        elif method == 'squared':
            num_sim_alg = _squared_sim

        elif method in ['exp', 'exponential']:
            num_sim_alg = _exp_sim

        elif method in ['gauss', 'gaussian']:
            num_sim_alg = _gauss_sim

        else:
            raise ValueError(
                "The algorithm '{}' is not known.".format(method))

        label = kwargs.pop('label', None)
        args = (num_sim_alg,) + args

        return self._compare_vectorized(
            _geo_internal, (lat1, lng1), (lat2, lng2), args, kwargs,
            label=label,
            name="geographic '{}'".format(method),
            description="Compare record pairs on geographic with '{}'"
                        "algorithm".format(method)
        )

    def date(self, s1, s2, *args, **kwargs):
        """
        date(self, s1, s2, swap_month_day=0.5, swap_months='default', missing_value=0, label=None)

        Compute the (partial) similarity between date values.

        Parameters
        ----------
        s1 : str or int
            The name or position of the column in the left DataFrame.
        s2 : str or int
            The name or position of the column in the right DataFrame.
        swap_month_day : float
            The value if the month and day are swapped.
        swap_months : list of tuples
            A list of tuples with common errors caused by the translating of
            months into numbers, i.e. October is month 10. The format of the
            tuples is (month_good, month_bad, value). Default : swap_months =
            [(6, 7, 0.5), (7, 6, 0.5), (9, 10, 0.5), (10, 9, 0.5)]
        missing_value : numpy.dtype
            The value for a comparison with a missing value. Default 0.
        label : label
            The label of the column in the resulting dataframe.
        """

        label = kwargs.pop('label', None)

        return self._compare_vectorized(
            _dates_internal, s1, s2, args, kwargs,
            label=label,
            name="date",
            description="Compare record pairs on dates."
        )


def _missing(*args):
    """ Missing values.

    Internal function to return the index of record pairs with missing values
    """

    return np.any(
        np.concatenate(
            [np.array(pandas.DataFrame(arg).isnull()) for arg in args],
            axis=1),
        axis=1)



import numpy


# Errors and Exception handlers
class IndexError(Exception):
    """ Error class for errors related to indexing. """
    pass


class VisibleDeprecationWarning(UserWarning):
    """Visible deprecation warning.
    Based on numpy's VisibleDeprecationWarning.
    """
    pass

# Checks and conversions


def is_label_dataframe(label, df):
    """check column label existance"""

    setdiff = set(label) - set(df.columns.tolist())

    if len(setdiff) == 0:
        return True
    else:
        return False


def get_length(x):
    """Return int or len(x)"""

    try:
        return int(x)
    except Exception:
        return len(x)


def listify(x):
    """Make a list of the argument if it is not a list."""

    if isinstance(x, list):
        return x
    elif isinstance(x, tuple):
        return list(x)
    else:
        return [x]


def unique(x):
    """Convert a list in a unique list."""

    return list(set(x))


def merge_dicts(*dict_args):
    '''
    Given any number of dicts, shallow copy and merge into a new dict,
    precedence goes to key value pairs in latter dicts.
    '''
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result


def split_index(index, chunks):
    """Function to split pandas.Index and pandas.MultiIndex objects.

    Split pandas.Index and pandas.MultiIndex objects into chunks.
    numpy.array_split returns incorrect results for MultiIndex
    objects. This function is based on numpy.split_array.

    Parameters
    ----------
    index : pandas.Index, pandas.MultiIndex
        A pandas.Index or pandas.MultiIndex to split into chunks.
    chunks : int
        The number of parts to split the index into.

    Returns
    -------
    list
        A list with chunked pandas.Index or pandas.MultiIndex objects.

    """

    Ntotal = index.shape[0]
    Nsections = int(chunks)
    if Nsections <= 0:
        raise ValueError('number sections must be larger than 0.')
    Neach_section, extras = divmod(Ntotal, Nsections)
    section_sizes = ([0] + extras * [Neach_section + 1] +
                     (Nsections - extras) * [Neach_section])
    div_points = numpy.array(section_sizes).cumsum()

    sub_ind = []
    for i in range(Nsections):
        st = div_points[i]
        end = div_points[i + 1]
        sub_ind.append(index[st:end])

    return sub_ind



import os

from setuptools import setup, find_packages
import versioneer


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


setup(
    name="recordlinkage",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    author="Jonathan de Bruin",
    author_email="jonathandebruinhome@gmail.com",

    platforms="any",

    # Description
    description="A record linkage toolkit for linking and deduplication",
    long_description=read('README.rst'),

    # Github
    url="https://github.com/J535D165/recordlinkage",

    install_requires=[
        "six>=1.10.0",
        "jellyfish>=0.5.4",
        "numpy>=1.9.0",
        "pandas>=0.18.0",
        "scipy>=0.17.1",
        "scikit-learn>=0.17.1",
    ],
    packages=find_packages(exclude=["recordlinkage._lib"]),
    include_package_data=True,
    package_dir={'recordlinkage': 'recordlinkage'},
    package_data={'recordlinkage': ['datasets/*/*.csv']},
    license='GPL-3.0'
)
